# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains the Poisson Bayes classifier."""
import os

import numpy as np
import pandas as pd
from scipy.stats import poisson
from tqdm import tqdm

from riid.sampleset import SampleSet


class PoissonBayes:
    """This class implements a Bayes classifier for poisson distributed data."""

    def __init__(self, seeds: SampleSet = None):
        """Initializes the classifier with seeds.

        This constructor takes "seed" spectra as input to be used as the shapes against which
        classifications will be performed. It will apply a PMF to the SampleSet.

        Args:
            seeds: an optional SampleSet of seed spectra for each potential source of interest.
                A background seed may be included if classification as background is desired.

        Returns:
            None

        Raises:
            NegativeSpectrumError: raised if any seed spectrum has negative counts in any bin.
            ZeroTotalCountsError: raised if any seed spectrum contains zero total counts.
        """
        self._temp_file_path = "temp.mdl"

        if not seeds:
            return

        if seeds.n_samples <= 0:
            raise ValueError("At least one seed must be provided.")
        if (seeds.spectra.values < 0).any():
            msg = "Argument 'seeds' can't contain any spectra with negative values."
            raise NegativeSpectrumError(msg)
        if (seeds.spectra.values.sum(axis=1) <= 0).any():
            msg = "Argument 'seeds' can't contain any spectra with zero total counts."
            raise ZeroTotalCountsError(msg)

        seeds.to_pmf()
        self._seeds = seeds.spectra
        self._classes = seeds.labels
        self._n_bins = self._seeds.shape[1]

    def _calculate_valid_bins(self, bg_spectrum: np.ndarray) -> np.ndarray:
        """Calculates which bins are valid for use in classification based on the values present in
        both the seeds and the given background spectrum.

        Args:
            bg_spectrum: a numpy.ndarray representing a single background spectrum.
                Shape: (n_bins,)

        Returns:
            A numpy.ndarray of booleans representing which channels are valid (True) and which
            are not (False).

        Raises:
            None
        """
        return ((self._seeds.values.T + bg_spectrum[:, None]) > 0).prod(axis=1).astype(bool)

    def _predict_single(self, seed_spectra: np.ndarray, gross_spectrum: np.ndarray,
                        bg_spectrum: np.ndarray, estimated_bg_counts: float,
                        normalize_scores: bool = False) -> np.ndarray:
        """Performs a prediction of the probabilities of each class for single spectrum.

        Performs prediction on the given spectrum and returns the probabilies of observing the
        spectrum from each of the underlying classes.
        Each probability represents the chances that you would have measured the the given spectrum
        with the underlying distribution of a particular seed.
        Note that these probabilities are almost always extremely low (which actually makes sense
        since all of this is technically very unlikely).

        Args:
            seed_spectra: a numpy.ndarray containing the seed spectra on which to match.
                1D shape: (n-seeds, n_bins).
            gross_spectrum: a numpy.ndarray containing the spectrum needing to be identified.
                1D shape: (n_bins,).
            bg_spectrum: a numpy.ndarray containing the background spectrum associated with the
                gross spectrum. 1D shape: (n_bins,).
            estimated_bg_counts: an estimate of the number of counts from background sources.
            normalize_scores: when True scores will be normalized based upon the highest
                achievable score (perfect template).

        Returns:
            Returns a numpy.ndarray contining the probabilities for each class.
            1D shape: (n_classes,)

        Raises:
            None
        """
        total_counts = gross_spectrum.sum()
        estimated_fgs = np.abs((total_counts - estimated_bg_counts) * seed_spectra)
        estimated_bg = estimated_bg_counts * bg_spectrum
        expected_signal = estimated_bg[None, :] + estimated_fgs
        if normalize_scores:
            expected_signal = np.vstack(
                (
                    expected_signal,
                    gross_spectrum,
                )
            )

        probas = poisson.logpmf(gross_spectrum, expected_signal).mean(axis=1)

        if normalize_scores:
            probas = probas[:-1] - probas[-1]

        return probas

    def predict(self, gross_ss: SampleSet, bg_ss: SampleSet,
                normalize_scores: bool = False, verbose=0) -> list:
        """Attempts to classify the provided gross spectra using the seeds with which the
        PoissonBayes object was originally initialized.

        Args:
            gross_ss: a SampleSet of `n` gross spectra where `n` >= 1.
            bg_ss: a SampleSet of one or `n` background spectra.
                If a SampleSet with one background spectrum is provided, that background will be
                applied to all gross spectra. Otherwise, the number of background spectra must
                equal the number of gross spectra. In the latter case, it is assumed that the
                index positions of corresponding background and gross spectra are the same.
            normalize_scores: when True scores will be normalized based upon the highest
                achievable score (perfect template).

        Returns:
            None

        Raises:
            ValueError: Raised when:
                - no gross spectra are provided.
                - no background(s) are provided.
                - an invalid number of backgrounds is provided.
                - the binning of the provided SampleSets do not match the binning of
                  the previously provided seed spectra.
        """
        if gross_ss.n_samples <= 0:
            raise ValueError("No gross spectr[a|um] provided!")
        if bg_ss.n_samples <= 0:
            raise ValueError("No background spectr[a|um] provided!")
        if bg_ss.n_samples != 1 and bg_ss.n_samples != gross_ss.n_samples:
            msg = "The number of background spectra is not 1 (it's {}).  ".format(bg_ss.n_samples)
            msg += "Therefore you must provide equal numbers of background and gross spectra!"
            raise ValueError(msg)
        if gross_ss.n_channels != self._n_bins:
            msg = "The binning of all gross spectra must be the same as the provided seeds!  "
            msg += "Seed spectra contain {} bins.".format(self._n_bins)
            raise ValueError(msg)
        if bg_ss.n_channels != self._n_bins:
            msg = "The binning of all background spectra must be the same as the provided seeds!  "
            msg += "Seed spectra all contain {} bins.".format(self._n_bins)
            raise ValueError(msg)

        bg_ss.to_pmf()
        probas = np.zeros([gross_ss.spectra.values.shape[0], self._classes.shape[0]])
        for i in tqdm(range(gross_ss.n_samples), desc="Samples", disable=not verbose, leave=False):
            lt_estimate = gross_ss.live_time.values[i]
            if i == 0 or bg_ss.n_samples > 1:
                bg_spectrum = bg_ss.spectra.values[i]
                valid_bins = self._calculate_valid_bins(bg_spectrum)
                seed_spectra = self._seeds.values[:, valid_bins]
                bg_cps = bg_ss.total_counts.values[i] / bg_ss.live_time.values[i]
            estimated_bg_counts = lt_estimate * bg_cps
            probas[i, :] = self._predict_single(
                seed_spectra,
                gross_ss.spectra.values[i][valid_bins],
                bg_spectrum[valid_bins],
                estimated_bg_counts,
                normalize_scores=normalize_scores
            )
        max_indices = probas.argmax(axis=1)
        gross_ss.predictions = self._classes[max_indices]
        gross_ss.prediction_probas = pd.DataFrame(probas, columns=self._classes)

    def save(self, file_path: str):
        """Saves the model to a file.

        Args:
            file_path: a string representing the file path at which to save the model.

        Returns:
            None

        Raises:
            ValueError: Raised when the given file path already exists.
        """
        if os.path.exists(file_path):
            raise ValueError("Path already exists.")
        self._seeds.to_hdf(file_path, "seeds")
        pd.DataFrame(self._classes).to_hdf(file_path, "classes")

    def load(self, file_path: str):
        """Loads the model from a file.

        Args:
            file_path: a string representing the file path from which to load the model.

        Returns:
            None

        Raises:
            None
        """
        self._seeds = pd.read_hdf(file_path, "seeds")
        self._classes = pd.read_hdf(file_path, "classes").values.flatten()
        self._n_bins = self._seeds.shape[1]

    def serialize(self) -> bytes:
        """Converts the model to a bytes object.

        Args:
            None

        Returns:
            Returns a bytes object representing the binary of an HDF file.

        Raises:
            None
        """
        self.save(self._temp_file_path)
        with open(self._temp_file_path, "rb") as f:
            data = f.read()
        os.remove(self._temp_file_path)
        return data

    def deserialize(self, stream: bytes):
        """Populates the current model with the give bytes object.

        Args:
            stream: a bytes object containing the model information.

        Returns:
            None

        Raises:
            None
        """
        with open(self._temp_file_path, "wb") as f:
            f.write(stream)
        self.load(self._temp_file_path)
        os.remove(self._temp_file_path)


class ZeroTotalCountsError(ValueError):
    pass


class NegativeSpectrumError(ValueError):
    pass



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module provides visualization functions primarily for visualizing SampleSets."""
import matplotlib
import matplotlib.pyplot as plt  # noqa: E402
import numpy as np
import pandas as pd
from matplotlib import cm, rcParams  # noqa: E402
from matplotlib.colors import ListedColormap
from numpy.random import choice
from scipy.special import comb
from seaborn import heatmap
from sklearn.metrics import confusion_matrix as confusion_matrix_sklearn
from sklearn.metrics import f1_score

from riid.sampleset import SampleSet

# DO NOT TOUCH what is set below, nor should you override them inside a function.
plt.style.use("default")
rcParams["font.family"] = "serif"
CM = cm.tab10
MARKER = "."


def save_or_show_plot(func):
    """Function decorator providing standardized handling of
    saving and/or showing matplotlib plots.

    Args:
        func: the function to call that builds the plot and
            returns a tuple of (Figure, Axes).
    """
    def save_or_show_plot_wrapper(*args, save_file_path=None, show=True,
                                  return_bytes=False, **kwargs):
        if return_bytes:
            matplotlib.use("Agg")
        fig, ax = func(*args, **kwargs)
        if save_file_path:
            fig.savefig(save_file_path)
        if show:
            plt.show()
        if save_file_path:
            plt.close(fig)
        if return_bytes:
            import io
            buf = io.BytesIO()
            fig.savefig(buf, format='png')
            buf.seek(0)
            plt.close(fig)
            return buf
        return fig, ax

    return save_or_show_plot_wrapper


@save_or_show_plot
def confusion_matrix(ss: SampleSet, as_percentage: bool = False, cmap: str = "binary",
                     title: str = None, value_format: str = None, value_fontsize: int = None,
                     figsize: str = None, alpha: float = None):
    """Generates a confusion matrix for a SampleSet.

    Prediction and label information is used to distinguish between correct
    and incorrect classifications using color (green for correct, red for incorrect).

    Args:
        ss: a SampleSet of events to plot.
        as_percentage: scales existing confusion matrix values to the range 0 to 100.
        cmap: the colormap to use for seaborn colormap function.
        title: the plot title.
        value_format: the format string controlling how values are displayed in the matrix cells.
        value_fontsize: the font size of the values displayed in the matrix cells.
        figsize: the figure size passed to the matplotlib subplots call.
        alpha: the degree of opacity.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        EmptyPredictionsArrayError: raised when the sampleset does not contain any predictions
    """
    y_true = ss.labels
    y_pred = ss.predictions
    labels = sorted(set(list(y_true) + list(y_pred)))

    if y_pred.size == 0:
        msg = "Predictions array was empty.  Have you called `model.predict(ss)`?"
        raise EmptyPredictionsArrayError(msg)

    if not cmap:
        cmap = ListedColormap(['white'])

    cm_values = confusion_matrix_sklearn(y_true, y_pred, labels=labels)
    if as_percentage:
        cm_values = np.array(cm_values)
        cm_values = cm_values / cm_values.sum(axis=1)
        if not value_format:
            value_format = ".1%"
    else:
        if not value_format:
            value_format = ".0f"

    heatmap_kwargs = {}
    if alpha:
        heatmap_kwargs.update({"alpha": alpha})
    if value_format:
        heatmap_kwargs.update({"fmt": value_format})
    if cmap:
        heatmap_kwargs.update({"cmap": cmap})

    fig, ax = plt.subplots(**{"figsize": figsize})
    mask = cm_values == 0
    ax = heatmap(cm_values, annot=True, linewidths=0.25, linecolor="grey", cbar=False,
                 mask=mask, **heatmap_kwargs)

    tick_locs = np.arange(len(labels)) + 0.5
    ax.set_ylabel("Truth")
    ax.set_yticks(tick_locs)
    ax.set_yticklabels(labels, rotation=0)

    ax.set_xlabel("Prediction")
    ax.set_xticks(tick_locs)
    ax.set_xticklabels(labels, rotation=90)

    ax.set_title(title)
    plt.tight_layout()
    return fig, ax


@save_or_show_plot
def plot_live_time_vs_snr(ss: SampleSet, overlay_ss: SampleSet = None, alpha: float = 0.5,
                          xscale: str = "linear", yscale: str = "log",
                          xlim: tuple = None, ylim: tuple = None,
                          title: str = "Live Time vs. SNR", sigma_line_value: float = None):
    """Plots SNR against live time for all samples in a SampleSet.

    Prediction and label information is used to distinguish between correct
    and incorrect classifications using color (green for correct, red for incorrect).

    Args:
        ss: a SampleSet of events to plot.
        overlay_ss: another SampleSet to color as black.
        alpha: the degree of opacity (not applied to overlay_ss scatterplot if used).
        xscale: the X-axis scale.
        yscale: the Y-axis scale.
        xlim: a tuple containing the X-axis min and max values.
        ylim: a tuple containing the Y-axis min and max values.
        title: the plot title.
        sigma_line_value: plots a sigma line representing the `value` number of
            standard deviations from background.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    correct_ss = ss.get_indices(ss.labels == ss.predictions)
    incorrect_ss = ss.get_indices(ss.labels != ss.predictions)
    if not xlim:
        xlim = (ss.live_time.min(), ss.live_time.max())
    if not ylim:
        if yscale == "log":
            ylim = (ss.snr_estimate.clip(1e-3).min(), ss.snr_estimate.max())
        else:
            ylim = (ss.snr_estimate.clip(0).min(), ss.snr_estimate.max())
    fig, ax = plt.subplots()

    ax.scatter(
        correct_ss.live_time,
        correct_ss.snr_estimate,
        c="green", alpha=alpha, marker=MARKER, label="Correct"
    )
    ax.scatter(
        incorrect_ss.live_time,
        incorrect_ss.snr_estimate,
        c="red", alpha=alpha, marker=MARKER, label="Incorrect"
    )
    if overlay_ss:
        plt.scatter(
            overlay_ss.live_time,
            overlay_ss.snr_estimate,
            c="black", marker="+", label="Event" + ("" if overlay_ss.n_samples == 1 else "s"),
            s=75
        )
    if sigma_line_value:
        live_times = np.linspace(xlim[0], xlim[1])
        background_cps = ss.collection_information["bg_counts_expected"][0] / ss.live_time[0]
        snrs = sigma_line_value / np.sqrt(live_times * background_cps)
        plt.plot(
            live_times,
            snrs,
            c="blue",
            alpha=alpha,
            label="{}-sigma".format(sigma_line_value),
            ls="dashed"
        )

    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_xlabel("Live Time (s)")
    ax.set_ylabel("Signal-to-Noise Ratio (SNR)")
    ax.set_title(title)
    fig.legend(loc="lower right")
    return fig, ax


@save_or_show_plot
def plot_strength_vs_score(ss: SampleSet, overlay_ss: SampleSet = None, alpha: float = 0.5,
                           marker_size=75, xscale: str = "log", yscale: str = "linear",
                           xlim: tuple = (None, None), ylim: tuple = (0, 1.05),
                           title: str = "Signal Strength vs. Score", sigma_line_value: float = None):
    """Plots strength against prediction score for all samples in a SampleSet.

    Prediction and label information is used to distinguish between correct
    and incorrect classifications using color (green for correct, red for incorrect).

    Args:
        ss: a SampleSet of events to plot.
        overlay_ss: another SampleSet to color as blue (correct) and/or black (incorrect).
        alpha: the degree of opacity (not applied to overlay_ss scatterplot if used).
        xscale: the X-axis scale.
        yscale: the Y-axis scale.
        xlim: a tuple containing the X-axis min and max values.
        ylim: a tuple containing the Y-axis min and max values.
        title: the plot title.
        sigma_line_value: plots a sigma line representing the `value` number of
            standard deviations from background.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    correct_ss = ss.get_indices(ss.labels == ss.predictions)
    incorrect_ss = ss.get_indices(ss.labels != ss.predictions)
    if not xlim:
        if xscale == "log":
            xlim = (ss.sigma.clip(1e-3).min(), ss.sigma.max())
        else:
            xlim = (ss.sigma.clip(0).min(), ss.sigma.max())
    fig, ax = plt.subplots()

    ax.scatter(
        correct_ss.sigma,
        correct_ss.prediction_probas.max(axis=1),
        c="green", alpha=alpha, marker=MARKER, label="Correct", s=marker_size
    )
    ax.scatter(
        incorrect_ss.sigma,
        incorrect_ss.prediction_probas.max(axis=1),
        c="red", alpha=alpha, marker=MARKER, label="Incorrect", s=marker_size
    )
    if overlay_ss:
        overlay_correct_ss = overlay_ss.get_indices(overlay_ss.labels == overlay_ss.predictions)
        overlay_incorrect_ss = overlay_ss.get_indices(overlay_ss.labels != overlay_ss.predictions)
        ax.scatter(
            overlay_correct_ss.sigma,
            overlay_correct_ss.prediction_probas.max(axis=1),
            c="blue", marker="*", label="Correct Event" + ("" if overlay_incorrect_ss.n_samples == 1 else "s"),
            s=marker_size*1.25
        )
        ax.scatter(
            overlay_incorrect_ss.sigma,
            overlay_incorrect_ss.prediction_probas.max(axis=1),
            c="black", marker="+", label="Incorrect Event" + ("" if overlay_incorrect_ss.n_samples == 1 else "s"),
            s=marker_size*1.25
        )
    if sigma_line_value:
        ax.vlines(
            sigma_line_value,
            0,
            1,
            colors="blue",
            alpha=alpha,
            label="{}-sigma".format(sigma_line_value),
            linestyles="dashed"
        )

    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_xlabel("Sigma (net / sqrt(background))")
    ax.set_ylabel("Score")
    ax.set_title(title)
    ax.legend()

    return fig, ax


@save_or_show_plot
def plot_snr_vs_score(ss: SampleSet, overlay_ss: SampleSet = None, alpha: float = 0.5,
                      marker_size=75, xscale: str = "log", yscale: str = "linear",
                      xlim: tuple = (None, None), ylim: tuple = (0, 1.05),
                      title: str = "SNR vs. Score"):
    """Plots SNR against prediction score for all samples in a SampleSet.

    Prediction and label information is used to distinguish between correct
    and incorrect classifications using color (green for correct, red for incorrect).

    Args:
        ss: a SampleSet of events to plot.
        overlay_ss: another SampleSet to color as blue (correct) and/or black (incorrect).
        alpha: the degree of opacity (not applied to overlay_ss scatterplot if used).
        xscale: the X-axis scale.
        yscale: the Y-axis scale.
        xlim: a tuple containing the X-axis min and max values.
        ylim: a tuple containing the Y-axis min and max values.
        title: the plot title.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    correct_ss = ss.get_indices(ss.labels == ss.predictions)
    incorrect_ss = ss.get_indices(ss.labels != ss.predictions)
    if not xlim:
        if xscale == "log":
            xlim = (ss.snr_estimate.clip(1e-3).min(), ss.snr_estimate.max())
        else:
            xlim = (ss.snr_estimate.clip(0).min(), ss.snr_estimate.max())
    fig, ax = plt.subplots()

    ax.scatter(
        correct_ss.snr_estimate,
        correct_ss.prediction_probas.max(axis=1),
        c="green", alpha=alpha, marker=MARKER, label="Correct", s=marker_size
    )
    ax.scatter(
        incorrect_ss.snr_estimate,
        incorrect_ss.prediction_probas.max(axis=1),
        c="red", alpha=alpha, marker=MARKER, label="Incorrect", s=marker_size
    )
    if overlay_ss:
        overlay_correct_ss = overlay_ss.get_indices(overlay_ss.labels == overlay_ss.predictions)
        overlay_incorrect_ss = overlay_ss.get_indices(overlay_ss.labels != overlay_ss.predictions)
        ax.scatter(
            overlay_correct_ss.snr_estimate,
            overlay_correct_ss.prediction_probas.max(axis=1),
            c="blue", marker="*", label="Correct Event" + ("" if overlay_correct_ss.n_samples == 1 else "s"),
            s=marker_size*1.25
        )
        ax.scatter(
            overlay_incorrect_ss.snr_estimate,
            overlay_incorrect_ss.prediction_probas.max(axis=1),
            c="black", marker="+", label="Incorrect Event" + ("" if overlay_incorrect_ss.n_samples == 1 else "s"),
            s=marker_size*1.25
        )

    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_xlabel("SNR (net / background)")
    ax.set_ylabel("Score")
    ax.set_title(title)
    ax.legend()

    return fig, ax


@save_or_show_plot
def plot_spectra(ss: SampleSet, is_in_energy: bool = False, limit: int = None,
                 figsize: tuple = None, xscale: str = "linear", yscale: str = "log",
                 xlim: tuple = (0, None), ylim: tuple = (1e-1, None),
                 ylabel: str = None, title: str = None, legend_loc: str = None) -> tuple:
    """Plots the spectra contained with a SampleSet.

    Args:
        ss: spectra to plot.
        is_in_energy: whether or not to try and use each spectrum's
            energy bin values to convert the spectrum from bins to energy.
        limit: the number of spectra to plot; None will plot all.
        figsize: the figure size passed to the matplotlib subplots call.
        xscale: the X-axis scale.
        yscale: the Y-axis scale.
        xlim: a tuple containing the X-axis min and max values.
        ylim: a tuple containing the Y-axis min and max values.
        ylabel: the Y-axis label.
        title: the plot title.
        legend_loc: the location in which to place the legend. Defaults to None.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        ValueError: is_in_energy=True but energy bin centers are missing for any spectra.
        ValueError: limit is not None and less than 1.
    """
    if is_in_energy and pd.isnull(ss.energy_bin_centers.reshape(-1)).any():
        msg = "When using 'is_in_energy' a valid energy calibration is required."
        raise ValueError(msg)
    if limit and limit < 1:
        raise ValueError("'limit' argument can not be less than 1.")

    if limit:
        ss = ss.get_indices(range(limit))
    fig, ax = plt.subplots(figsize=figsize)
    for i in range(len(ss.spectra.index)):
        label = ss.labels[i]

        if is_in_energy:
            xvals = ss.energy_bin_centers[i, :]
        else:
            xvals = np.arange(ss.n_channels)
        ax.plot(
            xvals,
            ss.spectra.iloc[i],
            label=label,
            color=CM(i),
        )

    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    if is_in_energy:
        if xscale == "log":
            ax.set_xlabel("log(Energy (keV))")
        else:
            ax.set_xlabel("Energy (keV)")
    else:
        if xscale == "log":
            ax.set_xlabel("log(Channel)")
        else:
            ax.set_xlabel("Channel")
    if ylabel:
        ax.set_ylabel(ylabel)
    elif yscale == "log":
        ax.set_ylabel("log(Counts)")
    else:
        ax.set_ylabel("Counts")
    if title:
        ax.set_title(title)
    else:
        ax.set_title("Gamma Spectr" + ("um" if ss.n_samples == 1 else "a"))
    if legend_loc:
        ax.legend(loc=legend_loc)
    else:
        ax.legend()

    return fig, ax


@save_or_show_plot
def plot_learning_curve(train_loss: list, validation_loss: list,
                        xscale: str = "linear", yscale: str = "linear",
                        xlim: tuple = (0, None), ylim: tuple = (0, None),
                        ylabel: str = "Loss", legend_loc: str = "upper right",
                        smooth: bool = False, title: str = None) -> tuple:
    """Plots training and validation loss curves.

    Args:
        train_loss: list of training loss values.
        validation_loss: list of validation loss values.
        xscale: the X-axis scale.
        yscale: the Y-axis scale.
        xlim: a tuple containing the X-axis min and max values.
        ylim: a tuple containing the Y-axis min and max values.
        smooth: whether or not to apply smoothing to the loss curves.
        title: the plot title.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        ValueError:
            - if either list of values is empty
    """
    train_loss = np.array(train_loss)
    validation_loss = np.array(validation_loss)
    if train_loss.size == 0:
        raise ValueError("List of training loss values was not provided.")
    if validation_loss.size == 0:
        raise ValueError("List of validation loss values was not provided.")

    if isinstance(train_loss[0], (list, tuple)):
        train_x = np.array([ep for ep, _ in train_loss])
        train_y = np.array([lv for _, lv in train_loss])
    else:
        train_x = np.arange(len(train_loss))
        train_y = np.array([lv for lv in train_loss])

    if isinstance(validation_loss[0], (list, tuple)):
        val_x = np.array([ep for ep, _ in validation_loss])
        val_y = np.array([lv for _, lv in validation_loss])
    else:
        val_x = np.arange(len(validation_loss))
        val_y = np.array([lv for lv in validation_loss])

    fig, ax = plt.subplots()
    if smooth:
        from scipy.interpolate import make_interp_spline

        # The 300 one the next line is the number of points to make between min and max
        train_xnew = np.linspace(train_x.min(), train_x.max(), 50)
        spl = make_interp_spline(train_x, train_y, k=3)
        train_ps = spl(train_xnew)

        val_xnew = np.linspace(val_x.min(), val_x.max(), 300)
        spl = make_interp_spline(val_x, val_y, k=3)
        val_ps = spl(val_xnew)

        ax.plot(train_xnew, train_ps, label="Train", color=CM(0))
        ax.plot(val_xnew, val_ps, label="Validation", color=CM(1))
        ax.hlines(train_ps[-1], xlim[0], train_x.max(), color=CM(0), linestyles="dashed")
        ax.hlines(val_ps[-1], xlim[0], val_x.max(), color=CM(1), linestyles="dashed")
    else:
        ax.plot(train_x, train_y, label="Train", color=CM(0))
        ax.plot(val_x, val_y, label="Validation", color=CM(1))
        ax.hlines(train_y[-1], xlim[0], val_x.max(), color=CM(0), linestyles="dashed")
        ax.hlines(val_y[-1], xlim[0], val_x.max(), color=CM(1), linestyles="dashed")
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.set_xlabel("Epoch")
    ax.set_ylabel(ylabel)
    if title:
        ax.set_title(title)
    else:
        ax.set_title("Learning Curve")
    ax.legend(loc=legend_loc)

    return fig, ax


@save_or_show_plot
def plot_count_rate_history(cr_history: list, sample_interval: float,
                            event_duration: float, pre_event_duration: float,
                            ylim: tuple = (0, None), title: str = None):
    """Plots a count rate history.

    Args:
        cr_history: list of count rate values.
        sample_interval: the time in seconds for which each count rate values was collected.
        event_duration: the time in seconds during which an anomalous source was present.
        pre_event_duration: the time in seconds at which the anomalous source appears
            (i.e., the start of the event).
        validation_loss: list of validation loss values.
        ylim: a tuple containing the Y-axis min and max values.
        title: the plot title.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    fig, ax = plt.subplots()

    time_steps = np.arange(
        start=-pre_event_duration,
        stop=len(cr_history) * sample_interval - pre_event_duration,
        step=sample_interval
    )
    ax.plot(
        time_steps,
        cr_history,
        color=CM(0)
    )
    ax.axvspan(
        xmin=0,
        xmax=event_duration,
        facecolor=CM(0),
        alpha=0.1
    )
    ax.set_ylim(ylim)
    ax.set_xlabel("Time (seconds)")
    ax.set_ylabel("Counts per second")
    if title:
        ax.set_title(title)
    else:
        ax.set_title("Count Rate History")

    return fig, ax


@save_or_show_plot
def plot_score_histogram(ss: SampleSet, yscale="log", ylim=(1e-1, None),
                         title="Score Distribution"):
    """Plots a histogram of all of the model prediction scores.

    Args:
        ss: SampleSet containing prediction_probas values.
        yscale: the Y-axis scale.
        ylim: a tuple containing the Y-axis min and max values.
        title: the plot title.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    fig, ax = plt.subplots()

    indices1 = ss.collection_information.index[ss.collection_information["sigma"] <= 5]
    values1 = ss.prediction_probas.loc[indices1].values.flatten()
    values1 = np.where(values1 > 0.0, values1, values1)
    indices2 = ss.collection_information.index[(ss.collection_information["sigma"] > 5) &
                                               (ss.collection_information["sigma"] <= 50)]
    values2 = ss.prediction_probas.loc[indices2].values.flatten()
    values2 = np.where(values2 > 0.0, values2, values2)
    indices3 = ss.collection_information.index[ss.collection_information["sigma"] > 50]
    values3 = ss.prediction_probas.loc[indices3].values.flatten()
    values3 = np.where(values3 > 0.0, values3, values3)

    width = 0.35
    bins = np.linspace(0.0, 1.0, 100)
    ax.bar(
        values3,
        bins,
        width,
        color="green"
    )
    ax.bar(
        values2,
        bins,
        width,
        color="yellow"
    )
    ax.bar(
        values1,
        bins,
        width,
        color="red"
    )

    ax.set_yscale(yscale)
    ax.set_ylim(ylim)
    ax.set_xlabel("Scores")
    ax.set_ylabel("Occurrences")
    ax.set_title(title)

    return fig, ax


@save_or_show_plot
def plot_n_isotopes_vs_f1_bayes(ss: SampleSet, seeds: SampleSet, title="Number of Isotopes vs. F1 Score"):
    """Plots the F1 score for different numbers of isotopes under consideration,
    specifically for a PoissonBayes model.

    Args:
        ss: SampleSet containing prediction_probas values.
        seeds: the same seeds that were used by the PoissonBayes model.
        title: the plot title.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    n_to_sample = 20
    isotopes = np.array(list(set(seeds.labels)))
    isotope_to_index = {isotope: [] for isotope in isotopes}

    for i in range(seeds.labels.shape[0]):
        isotope_to_index[seeds.labels[i]].append(i)

    f1_scores = np.zeros((isotopes.shape[0] - 1,))
    for n in range(1, isotopes.shape[0]):
        inds = [choice(isotopes.shape[0], n + 1, replace=False) for _ in
                range(min(n_to_sample, comb(isotopes.shape[0], n, exact=True)))]
        for j in inds:
            # the isotopes we are considering this iteration
            curr_isotopes = isotopes[j]
            assert len(j) == n + 1
            proba_indicies = []
            for iso in curr_isotopes:
                proba_indicies += isotope_to_index[iso]
            proba_indicies.sort()

            # get the isotopes, whose correct label is in the set of isotopes we are considering
            i_labels = range(ss.labels.shape[0])
            all_proba_indicies = [i for i in i_labels if ss.labels[i] in curr_isotopes]
            # get probas that we need
            curr_probas = ss.prediction_probas.values[all_proba_indicies]
            curr_probas = curr_probas[:, proba_indicies]
            max_indicies = curr_probas.argmax(axis=1)
            predictions = [seeds.labels[proba_indicies[i]] for i in max_indicies]
            labels = ss.labels[all_proba_indicies]
            f1_scores[n - 1] += (f1_score(labels, predictions, average="micro"))

        f1_scores[n - 1] /= len(inds)

    fig, ax = plt.subplots()
    plt.plot(np.arange(2, f1_scores.shape[0] + 2), f1_scores)
    ax.set_ylim((0, 1.1))
    ax.set_xlabel("Number of Isotopes")
    ax.set_ylabel("F1-Score")
    ax.set_title(title)
    return fig, ax


@save_or_show_plot
def plot_n_isotopes_vs_f1(f1_scores: list, title: str = "Number of Isotopes vs. F1 Score"):
    """Plots the pre-computed F1 score for different numbers of isotopes under consideration.

    Args:
        f1_scores: list of pre-computed F1 scores.
        title: the plot title.

    Returns:
        A tuple (Figure, Axes) of matplotlib objects.

    Raises:
        None
    """
    fig, ax = plt.subplots()
    ax.plot([x for x in range(1, len(f1_scores)+1)], f1_scores)
    ax.set_ylim((0, 1.1))
    ax.set_xlabel("Number of Isotopes")
    ax.set_ylabel("F1-Score")
    ax.set_title(title)
    return fig, ax


class EmptyPredictionsArrayError(Exception):
    pass



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains utilities for loading and saving SampleSet data files."""
import copy
import logging
import os
import pathlib
import pickle
import re
import warnings

import pandas as pd

from riid import DATA_PATH, DataDirectoryNotFoundError
from riid.sampleset import SampleSet


def _check_data_path():
    if not os.path.exists(DATA_PATH):
        raise DataDirectoryNotFoundError()


def check_iso_name(name: str):
    """ Validates whether or not the given string contains
        a properly formatted radioisotope name.

        Note that this function does NOT look up a string
        to determine if the string corresponds to a
        radioisotope that actually exists, it just checks
        the format.

        The regular expression used by this function
        looks for the following (in order):
            - 1 capital letter
            - 0 to 1 lowercase letters
            - 1 to 3 numbers
            - an optional "m" for metastable

        Examples of properly formatted isotope names:
          - Y88
          - Ba133
          - Ho166m

        Args:
          name: the string to be checked

        Returns:
          True if the string has a valid format,
          otherwise False.
    """
    validator = re.compile(r"^[A-Z]{1}[a-z]{0,1}[0-9]{1,3}m?$")
    other_valid_names = ["fiestaware"]
    match = validator.match(name)
    is_valid = match is not None or \
        name.lower() in other_valid_names
    return is_valid


def load_samples_from_file(file_path: str, verbose=1) -> SampleSet:
    """Load samples from the given file_path."""
    ss = None
    if os.path.isfile(file_path):
        message = "Found samples, loading '{}'.".format(file_path)
        if verbose:
            logging.info(message)
        try:
            raw_ss = read_hdf(file_path)
        except OSError:
            with open(file_path, "rb") as fin:
                raw_ss = pickle.load(fin)
        kwargs = raw_ss.__dict__
        # Make instance of most recent SampleSet using data from loaded file.
        ss = SampleSet(**kwargs)
    else:
        message = "No samples were found for the given parameters."
        if verbose:
            logging.info(message)
        ss = None
    return ss


def load_samples(detector: str, measured_or_synthetic: str, train_or_test: str, file_name: str,
                 verbose=1) -> SampleSet:
    """Load samples for a given detector, type, and file name.
    """
    file_path = os.path.join(DATA_PATH, detector, measured_or_synthetic, train_or_test, file_name)
    file_path = os.path.expanduser(file_path)
    ss = load_samples_from_file(file_path, verbose=verbose)
    return ss


def save_samples_to_file(ss: SampleSet, file_path: str, verbose=1):
    """Writes out the given sampleset to disk at the given path."""
    try:
        write_hdf(ss, file_path)
    except OSError:
        with open(file_path, "wb") as fout:
            pickle.dump(ss, fout)
    if verbose:
        logging.info(f"Saved SampleSet to '{file_path}'")


def save_samples(ss: SampleSet, file_name: str, detector: str = None,
                 measured_or_synthetic: str = None, purpose: str = None, verbose=1):
    """Save the given samples to the appropriate data directory location."""
    _check_data_path()

    if not ss:
        raise EmptySampleSetError("No samples were provided")

    if detector:
        ss.detector = detector
    if measured_or_synthetic:
        ss.measured_or_synthetic = measured_or_synthetic
    if purpose:
        ss.purpose = purpose

    output_dir = os.path.join(
        DATA_PATH,
        ss.detector,
        ss.measured_or_synthetic,
        ss.purpose
    )
    output_dir = os.path.expanduser(output_dir)

    pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    output_path = os.path.join(
        DATA_PATH,
        ss.detector,
        ss.measured_or_synthetic,
        ss.purpose,
        file_name
    )
    save_samples_to_file(ss, output_path, verbose)


def save_detector_model(model_contents: str, model_hash: str, detector_name: str):
    """Save detector model
    """
    output_fn = os.path.join(
        DATA_PATH,
        detector_name,
        f"{model_hash}.dat"
    )
    with open(output_fn, "w") as fout:
        fout.write(model_contents)


def load_seeds(detector: str, measured_or_synthetic: str, file_name: str, verbose=1) -> SampleSet:
    """Load seeds for a given detector, type, and file_name.
    """
    _check_data_path()

    load_path = os.path.join(DATA_PATH, detector, measured_or_synthetic, "seed", file_name)
    load_path = os.path.abspath(os.path.expanduser(load_path))

    if not os.path.isfile(load_path):
        message = "No seeds were found for the given configuration at path {}".format(load_path)
        raise NoSeedsFoundError(message)

    if verbose:
        message = "Found seeds, loading '{}'.".format(load_path)
        logging.info(message)

    try:
        raw_seeds = read_hdf(load_path)
    except OSError:
        raw_seeds = pickle.load(open(load_path, "rb"))

    ss = SampleSet(**raw_seeds.__dict__)
    return ss


def save_seeds(ss: SampleSet, file_name: str, detector: str = None,
               measured_or_synthetic: str = None, verbose=1):
    """Load seeds for a given detector, type, and file_name."""
    _check_data_path()

    if not ss:
        raise EmptySampleSetError("No seeds were provided.")

    if detector:
        ss.detector = detector
    if measured_or_synthetic:
        ss.measured_or_synthetic = measured_or_synthetic

    save_samples(ss, file_name, verbose=verbose)


def read_hdf(file_name: str) -> SampleSet:
    """ Reads sampleset class from hdf binary format."""
    spectra = pd.read_hdf(file_name, "spectra")
    try:
        collection_information = pd.read_hdf(file_name, "collection_information")
    except:
        collection_information = pd.read_hdf(file_name, "info")
    sources = pd.read_hdf(file_name, "sources")
    sources.columns = [i.split("___")[0] for i in sources.columns]
    features = pd.read_hdf(file_name, "features")
    prediction_probas = pd.read_hdf(file_name, "prediction_probas")
    other_info = pd.read_hdf(file_name, "other_info")
    kwargs = {
        "spectra": spectra,
        "collection_information": collection_information,
        "sources": sources,
        "features": features,
        "prediction_probas": prediction_probas,
    }
    for col in other_info.columns:
        kwargs.update({col: other_info.loc[0, col]})

    return SampleSet(**kwargs)


def write_hdf(ss: SampleSet, output_path: str):
    """ Writes sampleset class to hdf binary format."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        ss.spectra.to_hdf(output_path, "spectra")
        ss.collection_information.to_hdf(output_path, "collection_information")

        sources = copy.copy(ss.sources)
        s = sources.columns.to_series()
        sources.columns = s + "____" + s.groupby(s).cumcount().astype(str).replace({'0': ''})
        sources.to_hdf(output_path, "sources")
        pd.DataFrame(ss.features).to_hdf(output_path, "features")
        ss.prediction_probas.to_hdf(output_path, "prediction_probas")

        # Get desired properties
        properties = [
            "detector_hash",
            "neutron_detector_hash",
        ]
        other_info = {
            "detector": ss.detector,
            "config": ss.config,
            "sensor_information": ss.sensor_information,
            "measured_or_synthetic": ss.measured_or_synthetic,
            "subtract_background": ss.subtract_background,
            "purpose": ss.purpose,
            "comments": ss.comments,
            "predictions": ss.predictions,
            "energy_bin_centers": ss.energy_bin_centers,
            "energy_bin_edges": ss._energy_bin_edges,
        }
        for prop in properties:
            if prop in ss.__dict__:
                other_info.update({prop: ss.__dict__[prop]})
            else:
                print(f"Not actually considering {prop}")

        df_other_info = pd.DataFrame(data=[other_info])

        df_other_info.to_hdf(output_path, "other_info")


class NoSeedsFoundError(Exception):
    pass


class EmptySampleSetError(Exception):
    pass



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains utilities for dealing with list mode data."""
import datetime
import numpy as np
import pandas as pd


def get_list_file_spectra(list_file_path, sampling_frequency):
    """Returns the time varying histogram of the contents of a list file and the time edges."""
    data = read_list_file(list_file_path)["data"]
    return make_time_slice_histograms(data, sampling_frequency)


def make_time_slice_histograms(data, spectrum_rate_hz):
    """Takes dataframe with columns of timestamp and channel and a frequency in Hz.
       Outputs a tuple of the form: (2D array which is n_time_slices X n_channels, time_edges)
    """
    max_channel = _next_highest_power_of_two(data.channel.max())
    time_edges = np.arange(0, data.timestamp.max(), 1/spectrum_rate_hz)
    upper_indices = np.searchsorted(data.timestamp, time_edges[1:])
    lower_indices = np.insert(upper_indices[:-1], 0, 0)
    ranges = np.vstack((lower_indices, upper_indices)).T
    all_counts = data.channel.values
    hists = np.zeros([len(ranges), max_channel])
    for i, limits in enumerate(ranges):
        hists[i, :] = np.bincount(all_counts[limits[0]:limits[1]], minlength=max_channel)
    return hists, time_edges


def _next_highest_power_of_two(value):
    """Returns the next highest power of two for a given value (eg. given 938 will return 1024)."""
    p = 1
    lower = True
    while lower:
        nth_power = 2**p
        lower = (value > nth_power)
        p += 1

    return nth_power


def read_list_file(list_file_path):
    """Reads a list file and returns a dict with the keys "header", "times", and "amplitudes"."""
    data = np.fromfile(list_file_path, dtype="uint8")
    header_length = 256
    header_data = data[:header_length]

    data = np.fromfile(list_file_path, dtype="uint32")
    event_data = data[int(header_length/4):]

    header_data = _interpret_list_mode_header(header_data)
    assert (header_data["data_style"] == 1), "Only digiBASE data format has been implemented. \
            List data style was {}".format(header_data["data_style"])
    # Indicates digiBASE format
    times, amplitudes = _interpret_list_mode_data(event_data)
    data = pd.DataFrame(data=np.array(times/1e6, dtype=np.float64), columns=["timestamp"])
    data["channel"] = amplitudes
    offset = np.array(2**21/1e6, dtype=np.float64)
    jumps = np.logical_or(data.timestamp.diff() > 1e12, data.timestamp.diff() < 0)
    data["timestamp"] += jumps.cumsum() * offset

    return {"header": header_data, "data": data}


def _interpret_list_mode_header(header_bytes):
    """Reads in list mode header."""
    assert (len(header_bytes) == 256), "'header_bytes' must contain exactly 256 bytes"

    hdef = _get_header_definition()

    header_data = {}

    for key in hdef:
        target_format = hdef[key]["d_type"]
        begin, end = hdef[key]["offset"], hdef[key]["offset"] + hdef[key]["n_bytes"]
        bytes_of_interest = header_bytes[begin:end]
        if target_format == np.char:
            value = "".join(np.char.mod("%c", bytes_of_interest))
            if value == "\x01":
                value = ""
        else:
            value = np.frombuffer(bytes_of_interest, dtype=target_format)[0]
        if key == "acquisition_start_time":
            value = _ole_days_to_datetime(value)
        header_data.update({key: value})
    return header_data


def _determine_if_event(four_byte_int):
    """Checks if event bit is set."""
    return four_byte_int >> 31 == 0


def _interpret_list_mode_data(data_bytes):
    """Parses out bytes into different entries."""
    times, amplitudes = _interpret_event_word(data_bytes)
    event_entries = _determine_if_event(data_bytes)
    times = times[event_entries]
    amplitudes = amplitudes[event_entries]

    return times, amplitudes


def _ole_days_to_datetime(float_days):
    t0 = datetime.datetime(1899, 12, 30, 0, 0, 0)
    days = int(np.floor(float_days))
    d_rem = float_days-days
    hours = int(np.floor(d_rem*24))
    h_rem = (d_rem * 24) - hours
    minutes = int(np.floor(h_rem*60))
    m_rem = (h_rem * 60) - minutes
    seconds = int(np.floor(m_rem*60))
    s_rem = (m_rem * 60) - seconds
    micro_seconds = int(s_rem * 1e6)
    dt = t0 + datetime.timedelta(
        days=days,
        seconds=seconds,
        microseconds=micro_seconds,
        hours=hours,
        minutes=minutes
    )
    return dt


def _interpret_event_word(four_byte_int):
    """Interprets AN event word.
        BIT   | DESCRIPTION
        31    | 0 for event
        30-21 | Amplitude of Pulse
        20-0  | Time in microseconds that the event occured
    """
    amplitude_mask = 4292870144
    time_mask = 2097151

    amplitude = np.bitwise_and(amplitude_mask, four_byte_int) >> 21
    time = np.bitwise_and(time_mask, four_byte_int, dtype=np.uint64)

    return(time, amplitude)


def _interpret_time_stamp_word(four_byte_int):
    """Interprets an event word.
        BIT  | DESCRIPTION
        31   | 1 for time_only
        30-0 | Current Time in Microseconds
    """
    time_mask = 2147483647
    return np.bitwise_and(time_mask, four_byte_int)


def _get_header_definition():
    """Returns dictionary of header definition for list mode data."""
    definition = {
        "header_format": {
            "offset": 0,
            "n_bytes": 4,
            "d_type": np.int32
        },
        "data_style": {
            "offset": 4,
            "n_bytes": 4,
            "d_type": np.int32
        },
        "acquisition_start_time": {
            "offset": 8,
            "n_bytes": 8,
            "d_type": np.double
        },
        "device_address": {
            "offset": 16,
            "n_bytes": 80,
            "d_type": np.char
        },
        "mcb_type_string": {
            "offset": 96,
            "n_bytes": 9,
            "d_type": np.char
        },
        "device_serial_number": {
            "offset": 105,
            "n_bytes": 16,
            "d_type": np.char
        },
        "data_description": {
            "offset": 121,
            "n_bytes": 80,
            "d_type": np.char
        },
        "energy_calibration_valid": {
            "offset": 201,
            "n_bytes": 1,
            "d_type": np.char
        },
        "energy_units": {
            "offset": 202,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "e_cal_offset": {
            "offset": 206,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "e_cal_linear": {
            "offset": 210,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "e_cal_quadratic": {
            "offset": 214,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "shape_calibration_valid": {
            "offset": 218,
            "n_bytes": 1,
            "d_type": np.char
        },
        "s_cal_offset": {
            "offset": 219,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "s_cal_linear": {
            "offset": 223,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "s_cal_quadratic": {
            "offset": 227,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "conversion_gain": {
            "offset": 231,
            "n_bytes": 4,
            "d_type": np.int32
        },
        "detector_ID_number": {
            "offset": 235,
            "n_bytes": 4,
            "d_type": np.int32
        },
        "real_time": {
            "offset": 239,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "live_time": {
            "offset": 243,
            "n_bytes": 4,
            "d_type": np.float32
        },
        "unused": {
            "offset": 247,
            "n_bytes": 9,
            "d_type": np.char
        },
    }
    return definition



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains a multi-layer perceptron classifier."""
import os

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.models import Sequential, load_model
from tensorflow.keras.regularizers import l1, l2

from riid.models.metrics import multi_f1, single_f1
from riid.sampleset import SampleSet

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"


class MLPClassifier:
    def __init__(self, n_inputs: int, n_classes: int, hidden_layers: tuple = (512,),
                 activation: str = "relu", loss: str = "categorical_crossentropy", optimizer="adam",
                 metrics: list = ["accuracy", "categorical_crossentropy", multi_f1, single_f1],
                 l2_alpha: float = 1e-4, activity_regularizer = l1(0), dropout: float = 0.0,
                 load_model_path: str = None, labels: list = None):
        """Initializes the classifier.

        The model is implemented as a tf.keras.Sequential object.

        Args:
            n_inputs: the size of the input layer.
            n_classes: the number of unique labels.
            hidden_layers: a tuple defining the number and size of dense layers.
            activation: the activate function to use for each dense layer.
            loss: the loss function to use for training.
            optimizer: the optimizer to use for training.
            metrics: list of metrics to be evaluating during training.
            l2_alpha: the alpha value for the L2 regularization of each dense layer.
            activity_regularizer: the regularizer function applied to the output of each dense layer.
            dropout: the amount of dropout to apply to each dense layer.
            load_model_path: loads a model from disk, overriding all other parameters.
            labels: optionally store a friendly form of the labels.

        Returns:
            None

        Raises:
            None
        """
        if not load_model_path:
            self.model = Sequential()
            for layer, nodes in enumerate(hidden_layers):
                if layer == 0:
                    self.model.add(
                        Dense(
                            nodes,
                            input_shape=(n_inputs,),
                            activation=activation,
                            activity_regularizer=activity_regularizer,
                            kernel_regularizer=l2(l2_alpha),
                        )
                    )
                else:
                    self.model.add(
                        Dense(
                            nodes,
                            activation=activation,
                            activity_regularizer=activity_regularizer,
                            kernel_regularizer=l2(l2_alpha),
                        )
                    )
                if dropout > 0:
                    self.model.add(Dropout(dropout))

            self.model.add(Dense(n_classes, activation="softmax"))
            self.model.compile(loss=loss, optimizer=optimizer, metrics=metrics)
            self._classes = None
            self._seeds = None
            self.labels = labels
        else:
            self.load(load_model_path)

        self._temp_file_path = "temp.mdl"

    def fit(self, ss: SampleSet, batch_size: int = 200, epochs: int = 20, validation_split: float = 0.2,
            callbacks=None, val_ss: SampleSet = None, patience=15, es_monitor="val_loss", es_mode="min", es_verbose=0,
            verbose=0):
        """Fits a model to the given SampleSet.

        Args:
            ss: a SampleSet of `n` spectra where `n` >= 1.  The spectra in the
                SampleSet should all be normalized and/or pre-processed in the same way.
            batch_size: the number of samples per gradient update.
            epochs: maximum number of training iterations.
            validation_split: the percentage of the training data to use as validation data.
            callbacks: callbacks list to be passed to TensorFlow Model.fit() method.
            val_ss: manually provided validation set (instead of taking a portion of `ss`).
            patience: the number of epochs to wait for tf.keras.callbacks.EarlyStopping object.
            es_monitor: quantity to be monitored for tf.keras.callbacks.EarlyStopping object.
            es_mode: mode for tf.keras.callbacks.EarlyStopping object.
            es_verbose: verbosity level for tf.keras.callbacks.EarlyStopping object.

        Returns:
            A tf.History object.

        Raises:
            ValueError: Raised when:
                - no spectra are provided.
                - spectra channels do not match the size of the model inpute layer.
        """
        if ss.n_samples <= 0:
            raise ValueError("No spectr[a|um] provided!")
        if ss.n_channels != self.model.input_shape[1]:
            msg = "The binning of all spectra must be compatible with the model's input layer!  "
            msg += "Model input layer size is {}; SampleSet channels is {}.".format(self.model.input_shape[1], ss.n_channels)
            raise ValueError(msg)

        x_train = ss.get_features().astype(float)
        y_train = ss.label_matrix.values.astype(float)
        if val_ss:
            val_data = (
                val_ss.get_features().astype(float),
                val_ss.label_matrix.values.astype(float),
            )
            validation_split = None
        else:
            val_data = None
            row_order = np.random.shuffle(np.arange(x_train.shape[0]))
            # Enforce random validation split through shuffling
            x_train = x_train[row_order][0]
            y_train = y_train[row_order][0]

        es = EarlyStopping(
            monitor=es_monitor,
            patience=patience,
            verbose=es_verbose,
            restore_best_weights=True,
            mode=es_mode,
        )

        if callbacks:
            callbacks.append(es)
        else:
            callbacks = [es]

        history = self.model.fit(
            x_train,
            y_train,
            epochs=epochs,
            verbose=verbose,
            validation_split=validation_split,
            validation_data=val_data,
            callbacks=callbacks,
            shuffle=True,
            batch_size=batch_size,
        )
        self.labels = ss.label_matrix_labels
        return history

    def predict(self, ss: SampleSet):
        """Classifies the spectra in the provided SampleSet.
           Results are stored inside the SampleSet's prediction-related properties.

        Args:
            ss: a SampleSet of `n` spectra where `n` >= 1.  The spectra in the
                SampleSet should be normalized and/or pre-processing in the same
                way as spectra on which the model was previously trained.

        Returns:
            None

        Raises:
            None
        """
        x_test = ss.get_features().astype(float)
        results = self.model.predict(x_test)  # output size will be n_samples by n_classes
        ss.prediction_probas = pd.DataFrame(data=results, columns=self.labels)
        ss.predictions = self.labels[results.argmax(axis=1)]

    def evaluate(self, ss: SampleSet, verbose=0):
        """Evaluates the model on the provided SampleSet.

        This function is pretty much a wrapper around the tf.Model.evaluate()

        Args:
            ss: a SampleSet object on which to evaluate model performance

        Returns:
            A dictionary of results where keys are metric names and values
            are the metric values.

        Raises:
            None
        """
        x_test = ss.get_features().astype(float)
        y_true = ss.label_matrix.values.astype(float)
        scores = self.model.evaluate(x_test, y_true, verbose=verbose)
        results = {}
        for metric, score in zip(self.model.metrics_names, scores):
            results.update({metric: score})
        return results

    def to_tflite(self, file_path=None, quantize=False):
        """Converts the model to a TFLite model and
        optionally saves it to a file or quantizes it.

        Args:
            file_path: a string for the file path to which to save the model.
            quantize: a bool specifying whether or not to apply quantization to the TFLite model.

        Returns:
            None

        Raises:
            None
        """
        converter = tf.lite.TFLiteConverter.from_keras_model(self.model)
        if quantize:
            converter.optimizations = [tf.lite.Optimize.OPTIMIZE_FOR_SIZE]
        tflite_model = converter.convert()
        if file_path:
            open(file_path, "wb").write(tflite_model)
        return tflite_model

    def save(self, file_path: str):
        """Saves the model to a file.

        Args:
            file_path: a string representing the file path at which to save the model.

        Returns:
            None

        Raises:
            ValueError: Raised when the given file path already exists.
        """
        if os.path.exists(file_path):
            raise ValueError("Path already exists.")
        self.model.save(file_path, save_format="h5")
        pd.DataFrame(self.labels).to_hdf(file_path, "labels")
        pd.DataFrame(self._classes).to_hdf(file_path, "_classes")
        pd.DataFrame(self._seeds).to_hdf(file_path, "_seeds")

    def load(self, file_path: str):
        """Loads the model from a file.

        Args:
            file_path: a string representing the file path from which to load the model.

        Returns:
            None

        Raises:
            None
        """
        self.model = load_model(
            file_path,
            custom_objects={"multi_f1": multi_f1, "single_f1": single_f1}
        )
        self.labels = pd.read_hdf(file_path, "labels").values.flatten()
        try:
            self._classes = pd.read_hdf(file_path, "_classes").values.flatten()
        except KeyError:
            self._classes = None
        try:
            self._seeds = pd.read_hdf(file_path, "_seeds")
        except KeyError:
            self._seeds = None

    def serialize(self) -> bytes:
        """Converts the model to a bytes object.

        Args:
            None

        Returns:
            Returns a bytes object representing the binary of an HDF file.

        Raises:
            None
        """
        self.save(self._temp_file_path)
        with open(self._temp_file_path, "rb") as f:
            data = f.read()
        os.remove(self._temp_file_path)
        return data

    def deserialize(self, stream: bytes):
        """Populates the current model with the give bytes object.

        Args:
            stream: a bytes object containing the model information.

        Returns:
            None

        Raises:
            None
        """
        with open(self._temp_file_path, "wb") as f:
            f.write(stream)
        self.load(self._temp_file_path)
        os.remove(self._temp_file_path)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains custom loss functions."""
import numpy as np
from tensorflow.keras import backend as K


def negative_log_f1(y_true: np.ndarray, y_pred: np.ndarray):
    """Implements custom negative log F1 loss score for use in multi-isotope classifiers.

        Args:
            y_true: a list of ground truth.
            y_pred: a list of predictions to compare against the ground truth.

        Returns:
            Returns the custom loss score.

        Raises:
            None
    """
    diff = y_true - y_pred
    negs = K.clip(diff, -1.0, 0.0)
    false_positive = -K.sum(negs, axis=-1)
    true_positive = 1.0 - false_positive
    lower_clip = 1e-20
    true_positive = K.clip(true_positive, lower_clip, 1.0)

    return -K.mean(K.log(true_positive))


def negative_f1(y_true, y_pred) -:
    """Implements custom negative F1 loss score for use in multi-isotope classifiers.

        Args:
            y_true: a list of ground truth.
            y_pred: a list of predictions to compare against the ground truth.

        Returns:
            Returns the custom loss score.

        Raises:
            None
    """
    diff = y_true - y_pred
    negs = K.clip(diff, -1.0, 0.0)
    false_positive = -K.sum(negs, axis=-1)
    true_positive = 1.0 - false_positive
    lower_clip = 1e-20
    true_positive = K.clip(true_positive, lower_clip, 1.0)

    return -K.mean(true_positive)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains utilities for working with GADRAS and related files."""
import copy
import re
import struct
import subprocess
import tempfile
from collections import defaultdict

import numpy as np
import pandas as pd
from tqdm import tqdm

from riid.sampleset import SampleSet


# Define PCF definition per PCF ICD
HEADER_DEFINITIONS = defaultdict(lambda: {
    "fields": (
        "NRPS",
        "Version",
        "Ecal_label",
        "Energy_Calibration_Offset",
        "Energy_Calibration_Gain",
        "Energy_Calibration_Quadratic",
        "Energy_Calibration_Cubic",
        "Energy_Calibration_Low_Energy"
    ),
    "mask": "<h3s4cfffff",
    "n_bytes": [2, 3, 4, 4, 4, 4, 4, 4]
    }
)

HEADER_DEFINITIONS["DHS"] = {
    "fields": (
        "NRPS",
        "Version",
        "last_mod_hash",
        "UUID",
        "Inspection",
        "Lane_Number",
        "Measurement_Remark",
        "Intrument_Type",
        "Manufacturer",
        "Instrument_Model",
        "Instrument_ID",
        "Item_Description",
        "Item_Location",
        "Measurement_Coordinates",
        "Item_to_detector_distance",
        "Occupancy_Number",
        "Cargo_Type"
    ),
    "mask": "<h3s7s36s16sh26s28s28s18s18s20s16s16shh16s",
    "n_bytes": [2, 3, 7, 36, 16, 2, 26, 28, 28, 18, 18, 20, 16, 16, 2, 2, 16]
}

SPECTRUM_DEFINITION = {
    "fields": (
        "Compressed_Text_Buffer",
        "Date-time_VAX",
        "Tag",
        "Live_Time",
        "Total_time_per_real_time",
        "unused0",
        "unused1",
        "unused2",
        "Energy_Calibration_Offset",
        "Energy_Calibration_Gain",
        "Energy_Calibration_Quadratic",
        "Energy_Calibration_Cubic",
        "Energy_Calibration_Low_Energy",
        "Occupancy_Flag",
        "Total_Neutron_Counts",
        "Number_of_Channels"
    ),
    "mask": "180s23scffffffffffffi",
    "n_bytes": [180, 23, 1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
}


def pcf_to_smpl(pcf_file_path, verbose=False):
    """ Converts pcf file to sampleset object.

    Inputs:
        pcf_file_path: path to pcf file

    Returns:
        sampleset: object containing information of pcf file in sampleset format
    """
    return _pcf_dict_to_sampleset(_pcf_to_dict(pcf_file_path, verbose))


def smpl_to_pcf(ss: SampleSet, save_path, verbose=True):
    """ Saves a sampleset to a pcf file
    """
    return _dict_to_pcf(_smpl_to_dict(ss), save_path, verbose=verbose)


def _get_srsi(file_bytes):
    """ Obtains the PCF file Spectral Records Start Index (SRSI).

    Inputs:
        file_bytes: byte array of file contents

    Returns:
        return_value: String indicator for what Deviation pairs type is being used.
        srsi: Spectral Records Start Index.  Index in file where spectra records begin
        after skipping past deviation pairs.
    """
    test_lengths = [30, 20]
    index = 256
    return_value = "other"
    srsi = 2
    for i in test_lengths:
        value = struct.unpack(
            "{}s".format(i),
            file_bytes[index:index + i]
        )[0].decode("utf-8").strip()
        if value in "DeviationPairsInFile" or value in "DeviationPairsInFileCompressed":
            return_value = value
            srsi = 83
            break
    return(return_value, srsi)


def _get_spectrum_header_offset(spectrum_number, srsi, nrps):
    """ Calculates the PCF header offset for spectrum.

    Inputs:
        spectrum_number: number of spectrum to obtain offset of header for.
        srsi: Spectral record start index.
        nrps: number of records per spectrum

    Returns:
        offset of spectrum header
    """
    return 256*(srsi + nrps * (spectrum_number - 1) - 1)


def _read_header(header_bytes, header_def):
    """ Converts bytes of header using header definition.

    Inputs:
        header_bytes: byte array of values of header.
        header_def: Dictionary defining the: mask, field names, and lengths of each entry.

    Returns:
        header: dictionary containing the content of the header.
    """
    header_values = struct.unpack(header_def["mask"], header_bytes[:sum(header_def["n_bytes"])])
    header = {}
    for field, value in zip(header_def["fields"], header_values):
        if isinstance(value, bytes) and field != "last_mod_hash":
            value = value.decode("utf-8", "ignore")
        header.update({field: value})
    return header


def _read_spectra(data, n_rec_per_spec, spec_rec_start_indx, verbose=False):
    """ Reads spectra from PCF file.

    Inputs:
        data: byte array of pcf file contents.
        n_rec_per_spec: Defines the number of records per spectrum. (NRPS in pcf ICD)
        spec_rec_start_index: Defines spectral record start index. (SRSI in pcf ICD)
        verbose: Boolean flag which will enable or disable status updates of reading in spectra.

    Returns:
        spectra: list of dictionaries containing pcf file spectra information.
    """
    num_samples = int((len(data) - 256 * (spec_rec_start_indx - 1)) / (256 * n_rec_per_spec))
    spectra = []
    # The range below is due to the definition of the spectrum offset in
    # pcf definition document assuming a first index of 1
    progress_iterable = tqdm(
        range(1, num_samples + 1),
        leave=False,
        desc="Spectrum",
        disable=not(verbose)
    )
    for spectrum_number in progress_iterable:
        spectrum_header_offset = _get_spectrum_header_offset(
            spectrum_number,
            spec_rec_start_indx,
            n_rec_per_spec
        )
        header_def = SPECTRUM_DEFINITION
        spectrum_header = _read_header(
            data[spectrum_header_offset: spectrum_header_offset+256],
            header_def
        )

        spctrum_offset = spectrum_header_offset + 256
        n_channels = int(256*(n_rec_per_spec-1) / 4)
        values = struct.unpack(
            "{}f".format(n_channels),
            data[spctrum_offset: spctrum_offset+(4*n_channels)]
        )

        spectrum = {"header": spectrum_header, "spectrum": np.array(values)}
        spectra.append(spectrum)

    return spectra


def _pcf_to_dict(pcf_file_path, verbose=False):
    """ Converts .pcf file into a python dictionary

    returns a dict with keys "header" and "spectra"

    Header provides metadata about the collection.

    spectra is a list of "header" and "spectrum" pairs for each spectrum in file.
    """
    pcf_data = np.fromfile(pcf_file_path, dtype=np.uint8)

    version = struct.unpack("3s", pcf_data[2:5])[0].decode("utf-8")
    header_def = HEADER_DEFINITIONS[version]
    deviation_values = struct.unpack("5120f", pcf_data[512:512+20480])
    if not set(deviation_values) == {0} and verbose:
        print("Devitaion values exist in file")
    header = _read_header(pcf_data[:256], header_def)
    dev_type, spec_rec_start_indx = _get_srsi(pcf_data)
    header.update({"SRSI": spec_rec_start_indx, "DevType": dev_type})
    spectra = _read_spectra(pcf_data, header["NRPS"], spec_rec_start_indx, verbose=verbose)

    return {"header": header, "spectra": spectra}


def _pcf_dict_to_sampleset(pcf_dict):
    """ Converts pcf dictionary into a SampleSet
    """

    if pcf_dict["spectra"]:
        num_spectra = len(pcf_dict["spectra"])
        num_channels = len(pcf_dict["spectra"][0]["spectrum"])

        sources = []
        collection_information = []

        spectra = np.ndarray((num_spectra, num_channels))

        for i in range(0, num_spectra):
            foreground = pcf_dict["spectra"][i]
            f_text_buffer = foreground["header"]["Compressed_Text_Buffer"]
            fg_counts = sum(foreground["spectrum"])

            total_counts = sum(foreground["spectrum"])
            live_time = foreground["header"]["Live_Time"]

            distance_search = re.search("@ ([0-9,.]+)", f_text_buffer)
            if distance_search:
                distance = float(distance_search.group(1)) / 100
            else:
                distance = np.nan

            source_string_initial = f_text_buffer.split(",")[0]
            source_string_initial = source_string_initial.split("{")[0]

            an_finds = re.findall("an=([0-9]+)", f_text_buffer)
            if an_finds:
                an = int(an_finds[0])
            else:
                an = None
            ad_finds = re.findall("ad=([0-9, .]+)", f_text_buffer)
            if ad_finds:
                ad_string = ad_finds[0].replace(",", "")
                ad = float(ad_string)
            else:
                ad = None

            # if source_string_initial:
            #     if f_text_buffer[:4] == "1kgU":
            #         source_string_initial = f_text_buffer[4:]
            #     source = _format_isotope_name(source_string_initial)
            source = f_text_buffer.strip()
            if source == "":
                source = "background"

            # PCF file contains energy calibration terms which are defined as:
            # E_i = a0 + a1*x + a2*x^2 + a3*x^3 + a4 / (1 + 60*x)
            # where:
            #   a0 = order_0
            #   a1 = order_1
            #   a2 = order_2
            #   a3 = order_3
            #   a4 = low_E
            #   x = channel number
            #   E_i = Energy value of i"th channel

            order_0 = float(foreground["header"]["Energy_Calibration_Offset"])
            order_1 = float(foreground["header"]["Energy_Calibration_Gain"])
            order_2 = float(foreground["header"]["Energy_Calibration_Quadratic"])
            order_3 = float(foreground["header"]["Energy_Calibration_Cubic"])
            low_E = float(foreground["header"]["Energy_Calibration_Low_Energy"])

            info = {
                "live_time": live_time,
                "total_counts": total_counts,
                "fg_counts": fg_counts,
                "distance": distance,
                "atomic_number": an,
                "area_density": ad,
                "ecal_order_0": order_0,
                "ecal_order_1": order_1,
                "ecal_order_2": order_2,
                "ecal_order_3": order_3,
                "ecal_low_e": low_E,
                "date-time": foreground["header"]["Date-time_VAX"],
                "real_time": foreground["header"]["Total_time_per_real_time"],
                "occupancy_flag": foreground["header"]["Occupancy_Flag"],
                "tag": foreground["header"]["Tag"],
                "total_neutron_counts": foreground["header"]["Total_Neutron_Counts"],
                "descr": f_text_buffer,
                "count_rate": total_counts / live_time
            }

            collection_information.append(info)
            sources.append(source)
            spectra[i, :] = np.array(foreground["spectrum"])

        sources_df = pd.DataFrame()
        for source in set(sources):
            sources_df[source] = np.isin(sources, source)
        sources_df = sources_df[sources]
        sources_df["label"] = sources

        sample_set = SampleSet(
            spectra=pd.DataFrame(data=spectra),
            sources=sources_df,
            collection_information=pd.DataFrame(data=collection_information),
            measured_or_synthetic="synthetic",
            _pcf_metadata=pcf_dict["header"]
        )
        sample_set.energy_bin_centers = sample_set._get_energy_centers()
        # reorder sources table columns to match the order the seeds were read in

        return sample_set


def _format_isotope_name(name):
    """ places isotope in format of alpha followed by numeric (eg. Pu239 instead of 239Pu)
    """

    alpha_part = re.search("[a-z]+", name.lower())
    numeric_part = re.search("[0-9]+", name)

    if alpha_part and numeric_part:
        name = alpha_part.group(0).capitalize() + numeric_part.group(0)
    elif alpha_part:
        name = alpha_part.group(0).capitalize()
    elif numeric_part:
        name = numeric_part.group(0)
    else:
        name = "Unknown"
    return name


def _convert_header(header_dict, header_def):
    values = []
    for i, tar_len in zip(header_def["fields"], header_def["n_bytes"]):
        if "unused" not in i:
            value = header_dict[i]
            if isinstance(value, str):
                if len(value) < tar_len:
                    value = value.ljust(tar_len, " ")
                value = value.encode("utf-8")
            if i == "Tag":
                if value == b"":
                    value = b" "
            values.append(value)
        else:
            values.append(0.0)
    return struct.pack(header_def["mask"], *values)


def _spectrum_byte_offset(spectrum_index, n_records_per_spectrum, spec_rec_start_index=83):
    """ Gives byte offset in file for where spectrum should occur
    spectrum_index begins with index 1

    """
    return 256 * (spec_rec_start_index + n_records_per_spectrum * (spectrum_index - 1) - 1)


def _spectrum_to_bytes(spectrum):
    """ Converts spectrum to bytes
    """
    n_channels = len(spectrum)
    return struct.pack("{}f".format(n_channels), *spectrum)


def _dict_to_pcf(pcf_dict, save_path, verbose=True):
    """  Converts dictionary of pcf information into pcf file
    """
    header = pcf_dict["header"]
    n_records_per_spectrum = header["NRPS"]
    n_spectra = len(pcf_dict["spectra"])

    file_bytes = _convert_header(header, HEADER_DEFINITIONS["DHS"])
    file_bytes += struct.pack("20s", b"DeviationPairsInFile")
    loc_first_spectra = _spectrum_byte_offset(1, n_records_per_spectrum)
    n_pad = 512 - len(file_bytes)
    file_bytes += bytes((" " * n_pad).encode("utf-8"))
    n_pad = loc_first_spectra - len(file_bytes)
    file_bytes += bytes(n_pad)

    # Add the spectra to file
    for i in range(n_spectra):
        spectrum_dict = pcf_dict["spectra"][i]
        spectrum_header_dict = spectrum_dict["header"]
        spectrum_header_bytes = _convert_header(spectrum_header_dict, SPECTRUM_DEFINITION)
        spectrum_bytes = _spectrum_to_bytes(spectrum_dict["spectrum"])
        file_bytes += spectrum_header_bytes + spectrum_bytes
    # save binary file
    with open(save_path, "wb") as fout:
        fout.write(file_bytes)
    return file_bytes


def _smpl_to_dict(ss: SampleSet):
    n_samples = ss.n_samples
    n_channels = ss.n_channels
    n_records_per_spectrum = int((n_channels / 64) + 1)

    if ss._pcf_metadata:
        pcf_header = ss._pcf_metadata
    else:
        pcf_header = {
            "Cargo_Type": "                ",
            "DevType": "DeviationPairsInFile",
            "Inspection": "                ",
            "Instrument_ID": "                  ",
            "Instrument_Model": "                  ",
            "Intrument_Type": "                            ",
            "Item_Description": "                    ",
            "Item_Location": "                ",
            "Item_to_detector_distance": 0,
            "Lane_Number": 0,
            "Manufacturer": "                            ",
            "Measurement_Coordinates": "                ",
            "Measurement_Remark": "                          ",
            "NRPS": n_records_per_spectrum,
            "Occupancy_Number": 0,
            "SRSI": 83,
            "UUID": "                                    ",
            "Version": "DHS",
            "last_mod_hash": b"       "
        }

    spectra = []

    for i in range(n_samples):
        row, label = (ss.collection_information["descr"].fillna("").values[i], ss.labels[i])
        if row:
            descr = row
        else:
            descr = label

        header = {
            "Compressed_Text_Buffer": descr,
            "Energy_Calibration_Low_Energy": ss.ecal_low_e.fillna(0).iloc[i],
            "Energy_Calibration_Offset": ss.ecal_order_0.fillna(0).iloc[i],
            "Energy_Calibration_Gain": ss.ecal_order_1.fillna(0).iloc[i],
            "Energy_Calibration_Quadratic": ss.ecal_order_2.fillna(0).iloc[i],
            "Energy_Calibration_Cubic": ss.ecal_order_3.fillna(0).iloc[i],
            "Live_Time": ss.live_time.fillna(0).iloc[i],
            "Total_time_per_real_time": ss.collection_information["real_time"].fillna(0).iloc[i],
            "Number_of_Channels": int(n_channels),
            "Date-time_VAX": ss.collection_information["date-time"].fillna("").iloc[i],
            "Occupancy_Flag": ss.collection_information.occupancy_flag.fillna(0).iloc[i],
            "Tag": ss.collection_information.tag.fillna("").iloc[i],
            "Total_Neutron_Counts":
                ss.collection_information.total_neutron_counts.fillna(0).iloc[i],
        }

        spectrum = ss.spectra.values[i, :]
        spectra.append({"header": header, "spectrum": spectrum})
    return {"header": pcf_header, "spectra": spectra}


def _run_gadras_string_inject(path_to_detector, path_to_setup, output_path, verbose=True):
    PATH_TO_INJECT_EXE = "C:\\GADRAS\\Program\\InjectSource.exe"
    inputs = [PATH_TO_INJECT_EXE, path_to_detector, path_to_setup, output_path]

    output = subprocess.getoutput(inputs)
    if verbose:
        print(output)


def make_seeds(source_strings, detector_dir, detector_name, labels=None, tag=None):
    seed_pcf_path = tempfile.TemporaryFile()
    seed_pcf_path.name = "seeds.pcf"
    seed_list_path = tempfile.TemporaryFile()
    seed_list_path.name = "seeds_list.csv"

    column_titles = ["source", "real_time", "poisson", "distance", "height", "include_bg"]
    seeds = []
    for source_string in source_strings:
        seeds.append(
            {
                "source": source_string,
                "real_time": 1,
                "poisson": False,
                "distance": 5000,
                "height": 56,
                "include_bg": False,
            }
        )
    df = pd.DataFrame(seeds)
    df[column_titles].to_csv(seed_list_path.name, sep="\t", index=False, header=column_titles)

    # Generate seeds.pcf file
    _run_gadras_string_inject(detector_dir, seed_list_path.name, seed_pcf_path.name, True)

    seeds = pcf_to_smpl(seed_pcf_path, True)

    if labels:
        target_labels = copy.deepcopy(labels)
        seeds.labels = target_labels
        target_labels.append("label")
        seeds._sources.columns = target_labels

    seeds.purpose = "seed"
    seeds.measured_or_synthetic = "synthetic"
    seeds.to_pmf()

    return seeds



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This modules contains two classes for generating synthetic gamma spectra."""
import datetime
import random

import numpy as np
import pandas as pd

from riid.sampleset import SampleSet


class GammaSpectraSynthesizer():
    """Creates a set of synthetic gamma spectra."""
    _supported_purposes = ["train", "test"]
    _supported_functions = ["uniform", "log10", "discrete", "list"]

    def __init__(self, seeds: SampleSet, purpose: str = "train",
                 samples_per_seed: int = 100, background_cps: float = 300.0,
                 background_n_cps: float = 4,
                 subtract_background: bool = True, live_time_function: str = "uniform",
                 live_time_function_args=(0.25, 8.0), snr_function: str = "uniform",
                 snr_function_args=(0.01, 100.0), mixture_size: int = 1,
                 mixture_min_contribution: float = 0.1, random_state: int = None):
        """Constructs a synthetic gamma spectra generator.

        Arguments:
            seeds: the known distributions of counts across all channels such
                that all channels sum to 1. Must contain one seed labeled "background."
            purpose: the intended use of the SampleSet.  Recommended values: "train" or "test".
                "seed" is not a valid purpose for SampleSets generated by this class.
            samples_per_seed: the number of synthetic samples to randomly generate per seed.
            background_cps: the constant rate of gammas from background.
            background_n_cps: the constant rate of neutrons from background.
            subtract_background: if True, generated spectra are foreground-only.
                If False, generated spectra are gross spectra (foreground + background).
            live_time_function: the method of sampling for target live time values.
                Options: uniform; log10; discrete; list.
            live_time_function_args: the range of values which are sampled in the fashion
                specified by the `live_time_function` argument.
            snr_function: the method of sampling for target signal-to-noise ratio values.
                Options: uniform; log10; discrete; list.
            snr_function_args: the range of values which are sampled in the fashion
                specified by the `snr_function` argument.
            mixture_size: the number of seeds to mix together.
            mixture_min_contribution: the minimum ratio of counts that a seed must contribute.
            random_state: the random seed value used to reproduce specific data sets.
        """
        self.seeds = seeds
        self.purpose = purpose
        self.detector = seeds.detector
        self.samples_per_seed = samples_per_seed
        self.background_cps = background_cps
        self.background_n_cps = background_n_cps
        self.subtract_background = subtract_background
        self.live_time_function = live_time_function
        self.live_time_function_args = live_time_function_args
        self.snr_function = snr_function
        self.snr_function_args = snr_function_args
        self.mixture_size = mixture_size
        self.mixture_min_contribution = mixture_min_contribution
        self.random_state = random_state

    def __str__(self):
        output = "SyntheticGenerationConfig()\n"
        for k, v in sorted(vars(self).items()):
            output += "  {}: {}\n".format(k, str(v))
        return output[:-1]

    def __getitem__(self, key):
        item = getattr(self, key)
        return item

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def _get_distribution_values(self, function: str, function_args, n_values: int):
        if function not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(function))

        if function == "uniform":
            value = np.random.uniform(*function_args, size=n_values)
        elif function == "log10":
            log10_args = tuple(map(np.log10, function_args))
            value = np.power(10, np.random.uniform(*log10_args, size=n_values))
        elif function == "discrete":
            value = np.random.choice(function_args, size=n_values)
        elif function == "list":
            value = function_args

        return value

    def _get_live_time_targets(self, n_samples: int) -> list:
        """Obtains a list of random SNR target values."""
        return self._get_distribution_values(
            self.live_time_function,
            self.live_time_function_args,
            n_samples
        )

    def _get_snr_targets(self, n_samples: int) -> list:
        """Obtains a list of random SNR target values."""
        return self._get_distribution_values(
            self.snr_function,
            self.snr_function_args,
            n_samples
        )

    def _generate_mixture_seeds(self, fg_seeds, n_samples, n_per_gamma_rates):
        """Randomly mixes the given foreground seeds into new "mixture seeds".
        """
        mix_size = self.mixture_size
        mix_min = self.mixture_min_contribution
        mix_max = 1 - (mix_min * (mix_size - 1))
        MIX_TRIM = 1e-8
        n_seeds = fg_seeds.n_samples
        n_channels = fg_seeds.n_channels

        mix_indices = np.random.rand(n_samples, n_seeds).argsort(axis=1)[:, :mix_size]
        mix_ratios = np.random.rand(n_samples, mix_size).clip(mix_min+MIX_TRIM, mix_max-MIX_TRIM)
        mix_ratios = mix_ratios / mix_ratios.sum(axis=1)[:, None]
        while (mix_ratios < mix_min).sum():
            mix_ratios = mix_ratios.clip(mix_min+MIX_TRIM, mix_max-MIX_TRIM)
            mix_ratios = mix_ratios / mix_ratios.sum(axis=1)[:, None]
        mix_ratios.sort()

        source_matrix = np.zeros([n_samples, n_seeds])
        seed_combined = np.zeros([n_samples, n_channels])
        n_per_g_rate_combined = np.zeros(n_samples)
        for i, (seed, n_per_g_rate) in enumerate(zip(fg_seeds.spectra.values, n_per_gamma_rates)):
            seed = seed.clip(0)
            seed = seed / seed.sum()
            where = (mix_indices == i).astype(bool)
            rows = where.sum(axis=1).astype(bool)
            how_much = mix_ratios[where]
            n_per_g_source = how_much * n_per_g_rate
            contrib = seed * how_much[:, None]
            seed_combined[rows, :] += contrib
            source_matrix[rows, i] = how_much
            n_per_g_rate_combined[rows] += n_per_g_source
        n_per_g_rate_combined[np.isnan(n_per_g_rate_combined)] = 0

        return seed_combined, source_matrix, n_per_g_rate_combined

    def generate(self, verbose=0):
        """Generate a sample set of gamma spectra from the given config.

        Args:
            verbose: determines the verbosity of status messages.

        Returns:
            A SampleSet of synthetic gamma spectra.

        Raises:
            Exception: raised when the config is invalid.
            EmptySampleSetError: raised when the provided SampleSet is empty.
        """
        if self.random_state:
            random.seed(self.random_state)
            np.random.seed(self.random_state)

        isotope_seeds = self.seeds.get_indices(self.seeds.labels != "background")
        if "total_neutron_counts" in self.seeds.collection_information.columns:
            n_per_gamma_rates = self.seeds.collection_information.total_neutron_counts.values / \
                self.seeds.collection_information.fg_counts.values
        else:
            n_per_gamma_rates = np.zeros(isotope_seeds.n_samples)

        isotopes = isotope_seeds.labels
        background_seed = self.seeds.get_indices(self.seeds.labels == "background")
        first_only = np.full(background_seed.n_samples, False)
        first_only[0] = True
        background_seed = background_seed.get_indices(first_only)
        background_spectrum = background_seed.spectra.iloc[0, :].values
        background_spectrum_clipped = background_spectrum.clip(0)
        background_pdf = background_spectrum_clipped / background_spectrum_clipped.sum()
        n_samples = self.samples_per_seed * isotope_seeds.n_samples

        # Generate samples for each isotope seed
        lt_targets = self._get_live_time_targets(n_samples)
        snr_targets = self._get_snr_targets(n_samples)

        bg_counts_expected = lt_targets * self.background_cps
        fg_counts_expected = snr_targets * bg_counts_expected
        total_counts_expected = bg_counts_expected + fg_counts_expected
        fg_contrib = fg_counts_expected / total_counts_expected
        bg_contrib = bg_counts_expected / total_counts_expected

        seed_combined, source_matrix, n_per_gamma_rate_combined = self._generate_mixture_seeds(
            isotope_seeds,
            n_samples,
            n_per_gamma_rates
        )
        seed_combined = fg_contrib[:, None] * seed_combined
        seed_combined += bg_contrib[:, None] * background_seed.spectra.values[0, :]

        lam = seed_combined * total_counts_expected[:, None]
        gross_spectra = np.random.poisson(lam)
        total_counts = gross_spectra.sum(axis=1)

        bg_neutron_counts_expected = lt_targets * self.background_n_cps
        total_neutron_counts = np.random.poisson(
            total_counts * n_per_gamma_rate_combined + bg_neutron_counts_expected
        )
        excess_from_expected = total_counts - bg_counts_expected
        snr_estimates = excess_from_expected / bg_counts_expected
        sigmas = excess_from_expected / np.sqrt(bg_counts_expected)

        sources = pd.DataFrame(columns=isotopes, data=source_matrix)
        if self.mixture_size > 1:
            sources["label"] = "mixture"
        else:
            sources["label"] = [isotopes[i] for i in source_matrix.argmax(axis=1)]
        ecal = self.seeds.ecal_factors[:, 0]
        info = pd.DataFrame.from_dict({
            "live_time": lt_targets,
            "snr_target": snr_targets,
            "snr_estimate": snr_estimates,
            "bg_counts_expected": bg_counts_expected,
            "fg_counts_expected": fg_counts_expected,
            "total_counts": total_counts,
            "sigma": sigmas,
            "ecal_order_0": np.full(n_samples, ecal[0]),
            "ecal_order_1": np.full(n_samples, ecal[1]),
            "ecal_order_2": np.full(n_samples, ecal[2]),
            "ecal_order_3": np.full(n_samples, ecal[3]),
            "ecal_low_e": np.full(n_samples, ecal[4]),
            "real_time": lt_targets,
            "occupancy_flag": np.full(n_samples, 0),
            "total_neutron_counts": total_neutron_counts,
            "tag": np.full(n_samples, " "),
            "date-time": np.full(n_samples, datetime.datetime.now().strftime("%d-%b-%Y %H:%M:%S.00")),
            "descr": np.full(n_samples, None),
        })
        info.iloc[:, :15] = info.iloc[:, :15].astype(float)

        if self.subtract_background:
            spectra = gross_spectra - background_pdf[None, :] * \
                bg_counts_expected[:, None]
        else:
            spectra = gross_spectra

        ss = SampleSet(
            spectra=pd.DataFrame(spectra),
            collection_information=info,
            sources=sources
        )
        ss.purpose = self.purpose
        ss.measured_or_synthetic = "synthetic"
        ss.detector = self.detector
        ss.energy_bin_centers = ss._get_energy_centers()
        if self.mixture_size == 1:
            ss.relabel_to_max_source()

        return ss

    @property
    def seeds(self) -> SampleSet:
        """
        A SampleSet of non-poisson sampled gamma spectra representing the perfect responses given
        by a specific detector when observing a an isotope for an "sufficiently large" live time.
        Each seed generally represents a single source of radiation, such as K40, Th232, Ba133,
        Y88, etc., however other seeds which incorporate sources + shielding are perfectly valid.
        """
        return self._seeds

    @seeds.setter
    def seeds(self, value: SampleSet):
        if "background" not in value.labels:
            raise ValueError("A seed with the label 'background' must be provided.")
        self._seeds = value

    @property
    def purpose(self) -> str:
        """
        The intended purpose of the SampleSet being generated.
        Used to determine where in the data directory to auto-cache the sample set.
        """
        return self._purpose

    @purpose.setter
    def purpose(self, value: str):
        if value not in self._supported_purposes:
            msg = "Invalid purpose. Supported purposes are: {}".format(self._supported_purposes)
            raise ValueError(msg)
        self._purpose = value

    @property
    def detector(self) -> str:
        """
        The unique name of the physical detector associated with the provided seeds.
        Used to determine where in the data directory to auto-cache the sample set.
        """
        return self._detector

    @detector.setter
    def detector(self, value: str):
        self._detector = value

    @property
    def samples_per_seed(self):
        """The number of samples to create per seed (excluding the background seed)."""
        return self._samples_per_seed

    @samples_per_seed.setter
    def samples_per_seed(self, value: int):
        if not isinstance(value, int):
            raise TypeError("Property 'samples_per_seed' key must be of type 'int'!")

        self._samples_per_seed = value

    @property
    def background_cps(self) -> float:
        """Specifies the counts per second contributed by background radiation."""
        return self._background_cps

    @background_cps.setter
    def background_cps(self, value: float):
        self._background_cps = value

    @property
    def background_n_cps(self) -> float:
        """
        The neutron rate from background.
        """
        return self._background_n_cps

    @background_n_cps.setter
    def background_n_cps(self, value: float):
        self._background_n_cps = value

    @property
    def subtract_background(self) -> bool:
        """Specifies whether or not to include counts from background in the final spectra."""
        return self._subtract_background

    @subtract_background.setter
    def subtract_background(self, value: bool):
        self._subtract_background = value

    @property
    def live_time_function(self) -> str:
        """The function used to randomly sample the desired live time space."""
        return self._live_time_function

    @live_time_function.setter
    def live_time_function(self, value: str):
        if value not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(value))
        self._live_time_function = value

    @property
    def live_time_function_args(self) -> tuple:
        """The live time space to be randomly sampled."""
        return self._live_time_function_args

    @live_time_function_args.setter
    def live_time_function_args(self, value):
        self._live_time_function_args = value

    @property
    def snr_function(self) -> str:
        """The function used to randomly sample the desired signal-to-noise (SNR) ratio space."""
        return self._snr_function

    @snr_function.setter
    def snr_function(self, value: str):
        if value not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(value))
        self._snr_function = value

    @property
    def snr_function_args(self) -> tuple:
        """The signal-to-noise (SNR) space to be randomly sampled."""
        return self._snr_function_args

    @snr_function_args.setter
    def snr_function_args(self, value):
        self._snr_function_args = value

    @property
    def mixture_size(self) -> int:
        """The number of seeds to be used to generate a spectrum's foreground."""
        return self._mixture_size

    @mixture_size.setter
    def mixture_size(self, value: int):
        self._mixture_size = value

    @property
    def mixture_min_contribution(self) -> float:
        """
        The minimum percentage that a seed can contribute to a multi-isotope spectrum's total
        foreground counts.
        """
        return self._mixture_min_contribution

    @mixture_min_contribution.setter
    def mixture_min_contribution(self, value: float):
        self._mixture_min_contribution = value

    @property
    def random_state(self) -> int:
        """
        The seed for the random number generator.
        Used when trying to make reproducible SampleSets.
        """
        return self._random_state

    @random_state.setter
    def random_state(self, value: int):
        self._random_state = value


class PassbySynthesizer():
    """Creates synthetic pass-by events as sequences of gamma spectra."""
    _supported_functions = ["uniform", "log10", "discrete", "list"]

    def __init__(self, seeds: SampleSet, events_per_seed: int = 2, sample_interval: float = 0.125,
                 background_cps: float = 300.0, subtract_background: bool = True,
                 dwell_time_function: str = "uniform", dwell_time_function_args=(0.25, 8.0),
                 fwhm_function: str = "discrete", fwhm_function_args=(1,),
                 snr_function: str = "uniform", snr_function_args=(1.0, 10.0),
                 min_fraction: float = 0.005, random_state: int = None):
        """Constructs a synthetic passy-by generator.

        Arguments:
            seeds: the known distributions of counts across all channels such
                that all channels sum to 1. Must contain one seed labeled "background."
            events_per_seed: the number of events to create per seed.
            sample_interval: the live time of each sampled spectrum comprising the
                entire pass-by event.
            background_cps: the constant rate of gammas from background.
            subtract_background: if True, generated spectra are foreground-only.
                If False, generated spectra are gross spectra (foreground + background).
            dwell_time_function: the method of sampling for target background only time
                to occur after sample pass-by.
                Options: uniform; log10; discrete; list.
            dwell_time_function_args: the range of values which are sampled in the fashion
                specified by the `dwell_time_function` argument.
            fwhm_function: the method of sampling for target full width at half maximum
                for the pass-by event.
                Options: uniform; log10; discrete; list.
            fwhm_function_args: the range of values which are sampled in the fashion
                specified by the `fwhm_function` argument.
            snr_function: the method of sampling for target signal-to-noise ratio values.
                Options: uniform; log10; discrete; list.
            snr_function_args: the range of values which are sampled in the fashion
                specified by the `snr_function` argument.
            min_fraction: the SNR threshold, defined as the percentage of the peak, that
                signifies the beginning and end of the pass-by.
            random_state: the random seed value used to reproduce specific data sets.
        """
        self.seeds = seeds
        self.detector = seeds.detector
        self.events_per_seed = events_per_seed
        self.sample_interval = sample_interval
        self.background_cps = background_cps
        self.subtract_background = subtract_background
        self.dwell_time_function = dwell_time_function
        self.dwell_time_function_args = dwell_time_function_args
        self.fwhm_function = fwhm_function
        self.fwhm_function_args = fwhm_function_args
        self.snr_function = snr_function
        self.snr_function_args = snr_function_args
        self.min_fraction = min_fraction
        self.random_state = random_state

    def __str__(self):
        output = "SyntheticGenerationConfig()"
        for k, v in sorted(vars(self).items()):
            output += "  {}: {}".format(k, str(v))
        return output

    def __getitem__(self, key):
        item = getattr(self, key)
        return item

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def _get_distribution_values(self, function: str, function_args, n_values: int):
        if function not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(function))

        if function == "uniform":
            value = np.random.uniform(*function_args, size=n_values)
        elif function == "log10":
            log10_args = tuple(map(np.log10, function_args))
            value = np.power(10, np.random.uniform(*log10_args, size=n_values))
        elif function == "discrete":
            value = np.random.choice(function_args, size=n_values)
        elif function == "list":
            value = function_args

        return value

    def _get_dwell_time_targets(self, n_samples) -> list:
        """Obtains a list of random dwell time target values."""
        return self._get_distribution_values(
            self.dwell_time_function,
            self.dwell_time_function_args,
            n_samples
        )

    def _get_fwhm_targets(self, n_samples) -> list:
        """Obtains a list of random full-width-half-max (FWHM) target values."""
        return self._get_distribution_values(
            self.fwhm_function,
            self.fwhm_function_args,
            n_samples
        )

    def _get_snr_targets(self, n_samples) -> list:
        """Obtains a list of random SNR target values."""
        return self._get_distribution_values(
            self.snr_function,
            self.snr_function_args,
            n_samples
        )

    def _calculate_passby_shape(self, fwhm):
        """Returns a pass-by shape with maximum of 1 which goes from min_fraction to min_fraction of
        signal with specified fwhm.
        """
        lim = np.sqrt((1-self.min_fraction)/self.min_fraction)
        samples = np.arange(-lim, lim, self.sample_interval / fwhm / 2)
        return 1 / (np.power(samples, 2) + 1)

    def _generate_single_passby(self, fwhm, snr, dwell_time, seed_pdf, background_pdf, source):
        """Generates sampleset with a sequence of spectra representative of a single pass-by event.
        """
        event_snr_targets = self._calculate_passby_shape(fwhm) * snr
        n_event_spectra = len(event_snr_targets)
        dwell_targets = np.zeros(int(dwell_time / self.sample_interval))
        snr_targets = np.concatenate((event_snr_targets, dwell_targets))

        n_samples = len(snr_targets)
        live_times = np.ones(n_samples)

        bg_counts_expected = self.background_cps * live_times
        bg_spectra = np.random.poisson(background_pdf * bg_counts_expected[:, None])
        bg_counts = bg_spectra.sum(axis=1)

        fg_counts_expected = self.background_cps * snr_targets * live_times
        fg_spectra = np.random.poisson(seed_pdf * fg_counts_expected[:, None])
        fg_counts = fg_spectra.sum(axis=1)
        net_spectra = bg_spectra + fg_spectra
        total_counts = bg_counts + fg_counts
        excess_from_expected = total_counts - bg_counts_expected
        snr_estimates = excess_from_expected / bg_counts_expected
        sigmas = excess_from_expected / np.sqrt(bg_counts_expected)

        source_data = np.hstack(
            (np.full([n_samples, 1], 1), np.full([n_samples, 1], source))
        )
        sources = pd.DataFrame(columns=[source, "label"], data=source_data)

        collection_information = pd.DataFrame(
                data=np.vstack((
                    live_times,
                    snr_targets,
                    snr_estimates,
                    bg_counts,
                    fg_counts,
                    bg_counts_expected,
                    total_counts,
                    sigmas)
                ).T,
                columns=[
                    "live_time",
                    "snr_target",
                    "snr_estimate",
                    "bg_counts",
                    "fg_counts",
                    "bg_counts_expected",
                    "total_counts",
                    "sigma"
                ]
            )

        if self.subtract_background:
            spectra = net_spectra - bg_spectra
        else:
            spectra = net_spectra
        comments = {
            "fwhm": fwhm,
            "snr": snr,
            "dwell_time": dwell_time,
            "event_length": n_event_spectra,
            "total_length": n_samples,
            "source": source
        }
        ss = SampleSet(
            spectra=pd.DataFrame(spectra),
            collection_information=collection_information,
            sources=sources,
            comments=comments
        )
        ss.detector = self.detector
        ss.purpose = "passby"
        ss.measured_or_synthetic = "synthetic"
        ss.energy_bin_centers = ss._get_energy_centers()

        return ss

    def generate(self, verbose=0):
        """Generate a list of sample sets where each represents a pass-by event.

        Args:
            verbose: determines the verbosity of status messages.

        Returns:
            A list of SampleSets where each SampleSet represents a pass-by event.

        Raises:
            None
        """
        if self.random_state:
            random.seed(self.random_state)
            np.random.seed(self.random_state)

        isotope_seeds = self.seeds.get_indices(self.seeds.labels != "background")
        background_seed = self.seeds.get_indices(self.seeds.labels == "background")
        bs_indices = np.full(background_seed.n_samples, False)
        bs_indices[0] = True
        background_seed = background_seed.get_indices(bs_indices)
        background_spectrum = background_seed.spectra.iloc[0, :].values
        background_pdf = background_spectrum.clip(0)
        background_pdf = background_pdf / background_pdf.sum()

        # Generate samples for each seed
        args = []
        for i, _ in enumerate(isotope_seeds.spectra.index):
            seed_spectrum = isotope_seeds.spectra.iloc[i, :].values
            source = isotope_seeds.labels[i]
            seed_pdf = seed_spectrum.clip(0)
            seed_pdf = np.array(seed_pdf) / seed_pdf.sum()
            n_samples = self.events_per_seed
            if isinstance(self.events_per_seed, dict):
                n_samples = self.events_per_seed[source]
            fwhm_targets = self._get_fwhm_targets(n_samples)
            snr_targets = self._get_snr_targets(n_samples)
            dwell_time_targets = self._get_dwell_time_targets(n_samples)
            for fwhm, snr, dwell in zip(fwhm_targets, snr_targets, dwell_time_targets):
                args.append((fwhm, snr, dwell, seed_pdf, background_pdf, source))

        events = [self._generate_single_passby(*a) for a in args]
        return events

    @property
    def seeds(self) -> SampleSet:
        """
        A SampleSet of non-poisson sampled gamma spectra representing the perfect responses given
        by a specific detector when observing a an isotope for an "sufficiently large" live time.
        Each seed generally represents a single source of radiation, such as K40, Th232, Ba133,
        Y88, etc., however other seeds which incorporate sources + shielding are perfectly valid.
        """
        return self._seeds

    @seeds.setter
    def seeds(self, value: SampleSet):
        if "background" not in value.source_types:
            raise ValueError("A seed with the label 'background' must be provided.")
        self._seeds = value

    @property
    def detector(self) -> str:
        """
        The unique name of the physical detector associated with the provided seeds.
        Used to determine where in the data directory to auto-cache the sample set.
        """
        return self._detector

    @detector.setter
    def detector(self, value: str):
        self._detector = value

    @property
    def sample_interval(self) -> float:
        """The sample interval (in seconds) at which the events are simulated."""
        return self._sample_interval

    @sample_interval.setter
    def sample_interval(self, value: float):
        self._sample_interval = value

    @property
    def events_per_seed(self):
        """The number of samples to create per seed (excluding the background seed)."""
        return self._events_per_seed

    @events_per_seed.setter
    def events_per_seed(self, value):
        if not isinstance(value, int) and not isinstance(value, dict):
            raise TypeError("Property 'events_per_seed' key must be of type 'int' or 'dict'!")

        self._events_per_seed = value

    @property
    def background_cps(self) -> float:
        """Specifies the counts per second contributed by background radiation."""
        return self._background_cps

    @background_cps.setter
    def background_cps(self, value: float):
        self._background_cps = value

    @property
    def subtract_background(self) -> bool:
        """Specifies whether or not to include counts from background in the final spectra."""
        return self._subtract_background

    @subtract_background.setter
    def subtract_background(self, value: bool):
        self._subtract_background = value

    @property
    def dwell_time_function(self) -> str:
        """The function used to randomly sample the desired dwell time space."""
        return self._dwell_time_function

    @dwell_time_function.setter
    def dwell_time_function(self, value: str):
        if value not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(value))
        self._dwell_time_function = value

    @property
    def dwell_time_function_args(self) -> tuple:
        """The dwell time space to be randomly sampled."""
        return self._dwell_time_function_args

    @dwell_time_function_args.setter
    def dwell_time_function_args(self, value):
        self._dwell_time_function_args = value

    @property
    def fwhm_function(self) -> str:
        """The function used to randomly sample the desired full-width-half-max (FWHM) ratio space.
        """
        return self._fwhm_function

    @fwhm_function.setter
    def fwhm_function(self, value: str):
        if value not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(value))
        self._fwhm_function = value

    @property
    def fwhm_function_args(self) -> tuple:
        """The full-width-half-max (FWHM) space to be randomly sampled."""
        return self._fwhm_function_args

    @fwhm_function_args.setter
    def fwhm_function_args(self, value):
        self._fwhm_function_args = value

    @property
    def snr_function(self) -> str:
        """The function used to randomly sample the desired signal-to-noise (SNR) ratio space."""
        return self._snr_function

    @snr_function.setter
    def snr_function(self, value: str):
        if value not in self._supported_functions:
            raise ValueError("{} is not a valid function.".format(value))
        self._snr_function = value

    @property
    def snr_function_args(self) -> tuple:
        """The signal-to-noise (SNR) space to be randomly sampled."""
        return self._snr_function_args

    @snr_function_args.setter
    def snr_function_args(self, value):
        self._snr_function_args = value

    @property
    def min_fraction(self) -> float:
        """Specifies the percentage of the peak amplitude to exclude."""
        return self._min_fraction

    @min_fraction.setter
    def min_fraction(self, value: float):
        self._min_fraction = value

    @property
    def random_state(self) -> int:
        """
        The seed for the random number generator.
        Used when trying to make reproducible SampleSets.
        """
        return self._random_state

    @random_state.setter
    def random_state(self, value: int):
        self._random_state = value



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module contains the SampleSet class."""
import copy
import logging
import random

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d


class SampleSet():
    """Container for collection of samples.
    """
    # pylint: disable=R0902
    # pylint: disable=R0904

    def __init__(self, **kwargs):
        # Give all class attributes some default values
        self._spectra = pd.DataFrame()  # expected to be of size n_samples X n_channels
        # expected to be of size n_samples X (n_source_types + 1) Labels Column
        # will exist
        self._sources = pd.DataFrame()
        # expected to be of size n_samples X (n feature columns defined)
        self._features = pd.DataFrame()
        self._detector = None
        df_columns = [
            "live_time",
            "snr_target",
            "snr_estimate",
            "bg_counts",
            "fg_counts",
            "bg_counts_expected",
            "total_counts",
            "sigma",
            "distance",
            "atomic_number",
            "area_density",
            "ecal_order_0",
            "ecal_order_1",
            "ecal_order_2",
            "ecal_order_3",
            "ecal_low_e",
            "date-time",
            "real_time",
            "occupancy_flag",
            "tag",
            "total_neutron_counts",
            "descr"
        ]
        self._collection_information = pd.DataFrame(columns=df_columns)
        self._sensor_information = {}
        self._config = None
        self._prediction_probas = pd.DataFrame(np.array([]))
        self._predictions = np.array([])
        self._measured_or_synthetic = None
        self._subtract_background = None
        self._purpose = None
        self._comments = ""
        self._energy_bin_centers = np.array([])
        self._energy_bin_edges = np.array([])
        self._pcf_metadata = None
        self.detector_hash = ""
        self.neutron_detector_hash = ""

        # Populate values passed in.
        if kwargs:
            for key, value in kwargs.items():
                self[key] = value

        # Check if updated values were missing fields.
        missing_columns = [column for column in df_columns if
                           column not in self._collection_information.columns]
        if missing_columns:
            self._collection_information = pd.concat(
                    (
                        self._collection_information,
                        pd.DataFrame(columns=missing_columns)
                    ),
                    sort=False
            )

    def __str__(self):
        return "Sampleset for {} containing {} samples with {} channels.".format(
            self.detector,
            self.n_samples,
            self.n_channels
        )

    def __getitem__(self, key):
        item = None
        try:
            if key == "sources":
                item = self.sources.replace(np.nan, 0)
            else:
                item = getattr(self, key)
        except KeyError as error:
            msg = "{0}".format(error)
            print(msg)
        return item

    def __setitem__(self, key, value):
        try:
            setattr(self, key, value)
        except KeyError as error:
            msg = "{0}".format(error)
            print(msg)

    def concat(self, ss_list):
        """ Provides method of combining many samplesets quickly
        """
        if not isinstance(ss_list, list):
            ss_list = [ss_list]

        all_spectra = [self.spectra]
        for ss in ss_list:
            spectra = ss.spectra
            all_spectra.append(spectra)
        self._spectra = pd.concat(all_spectra, ignore_index=True, sort=False)

        orig_sources = self.label_matrix
        orig_sources.loc[:, "label"] = self.labels
        all_sources = [orig_sources]
        for ss in ss_list:
            sources = ss.label_matrix
            sources.loc[:, "label"] = ss.labels
            all_sources.append(sources)

        self._sources = pd.concat(all_sources, ignore_index=True, sort=False)
        self._sources.replace(np.nan, 0.0, inplace=True)

        all_collection_information = [self._collection_information]
        for ss in ss_list:
            all_collection_information.append(ss._collection_information)
        self._collection_information = pd.concat(all_collection_information, ignore_index=True, sort=False)

        self._prediction_probas = pd.DataFrame(np.array([]))
        self._predictions = np.array([])

    def to_pmf(self, clip_negatives=True):
        """ Converts spectra to probability mass function (PMF)
        """
        if clip_negatives:
            self._spectra = self._spectra.clip(0)

        if (self._spectra.values < 0).sum():
            logging.warning(
                "You are performing a PMF operation on spectra which contains negative values.  " +
                "Consider applying `clip_negatives`."
            )

        self._spectra = pd.DataFrame(
            data=self._spectra.values / self._spectra.values.sum(axis=1)[:, np.newaxis]
        )

    def pmf_to_counts(self, verbose=0):
        """Converts spectra from PMF to counts.
        """
        row_sums = self._spectra.values.sum(axis=1)
        rows_eq_one = np.where(abs(row_sums - 1.0) <= 1e-5)[0]
        rows_neq_one = np.where(abs(row_sums - 1.0) > 1e-5)[0]
        for idx in rows_eq_one:
            self._spectra.iloc[idx] = \
                self._spectra.iloc[idx] * self._collection_information.loc[idx, "total_counts"]

        logging.info("PMF to Counts Conversion Summary:")
        logging.info(f"  Spectra converted:         {len(rows_eq_one)}")
        if rows_neq_one.size != 0:
            logging.info(f"  Spectra not converted:     {rows_neq_one.size}")
            if verbose:
                logging.info(f"      {rows_neq_one}")
                logging.info(f"      {row_sums[rows_neq_one]}")

    def to_count_rate(self):
        """ Converts spectra to count rage (live_time normalized)
        """
        self._spectra = pd.DataFrame(
            data=self._spectra.values / self.live_time[:, np.newaxis])

    def clip_negatives(self, min_value=0):
        """ Sets negative values to min_value
        """
        self._spectra = pd.DataFrame(data=self._spectra.clip(min_value))

    def replace_na(self, replace_value=0):
        """ Replaces np.nan() values with replace_value.
        """
        self._spectra.replace(np.nan, replace_value)
        self._sources.replace(np.nan, replace_value)
        self._collection_information.replace(np.nan, replace_value)

    def downsample_spectra(self, target_bins=128, binning_type="uniform"):
        """ Replaces spectra with downsampled version.  Uniform binning is assumed.
        """
        if binning_type == "uniform":
            if self.n_channels % target_bins == 0:
                n_per_group = int(self.n_channels / target_bins)
                transformation = np.zeros([target_bins, self.n_channels])
                for i in range(target_bins):
                    transformation[i, (i * n_per_group):((i * n_per_group) + n_per_group)] = 1
            else:
                msg = "Current binning ({}) is not a multiple of the target binning ({})".format(
                    self.n_channels,
                    target_bins
                )
                raise ValueError(msg)
            self._spectra = pd.DataFrame(
                data=np.matmul(
                    self._spectra.values,
                    transformation.T))
            self._energy_bin_centers = self._get_energy_centers()
        elif binning_type == "sqrt":
            n_source = self.n_channels

            max_n_per = (n_source / target_bins - 1) * 2 + 1
            n_per = np.round(np.linspace(1, max_n_per, target_bins))
            n_off = int(n_source - n_per.sum())

            if n_off > 0:
                n_per[-n_off:] += 1
            elif n_off < 0:
                n_per[n_off:] -= 1
            n_per = n_per.clip(1)

            if n_per.sum() != n_source:
                raise RebinningCalculationError("Rebinning calculation implemented incorrectly.")

            n_per = n_per.astype(int)
            transformation = np.zeros([n_source, target_bins])
            i_start = 0
            for col, number_of_channels in enumerate(n_per):
                i_finish = i_start + number_of_channels
                transformation[i_start:i_finish, col] = 1
                i_start = i_finish
            self._spectra = pd.DataFrame(
                data=np.matmul(
                    self._spectra.values,
                    transformation))
            self._energy_bin_centers = self._get_energy_centers()
        else:
            msg = "Binning must be 'uniform' or 'sqrt', '{}' was entered.".format(binning_type)
            raise ValueError(msg)

    def upsample_spectra(self, target_bins=4096):
        """ Replaces spectra with upsampled version.  Uniform binning is assumed.
        """
        if target_bins % self.n_channels == 0:
            n_per_group = int(target_bins / self.n_channels)
            transformation = np.zeros([target_bins, self.n_channels])
            for i in range(self.n_channels):
                transformation[(i * n_per_group):((i * n_per_group) + n_per_group), i] = \
                                    1.0 / n_per_group
        else:
            raise ValueError(
                ("Desired number of bins ({}) is cannot"
                 " be uniformly mapped from existing"
                 " channels ({})").format(target_bins, self.n_channels))
        self._spectra = pd.DataFrame(
            data=np.matmul(
                self._spectra.values,
                transformation.T))

    def relabel_from_dict(self, mapping_dict, mode="both"):
        """ Relabels labels and predictions according to "mapping_dict".
        "mapping_dict" should have "source":"target_label" pairs.

        mode="all: performs relableing of "label" and "label_matrix_label"
        mode="single": performs relabeling of "label" column of "sources" DataFrame
        mode="multi": performs relabeling of label_matrix columns of "sources" DataFrame
        """

        if mode not in ["both", "single", "multi"]:
            raise ValueError("Mode " + mode + " must be in set {'all', 'single', 'multi'}.")

        if mode in ["both", "single"]:
            for label in set(self.labels):
                if label not in mapping_dict.keys():
                    msg = str(label)
                    msg += " from single label was not specified in mapping dictionary."
                    raise MissingKeyForRelabeling(msg)
            orig_single_labels = self.labels
            self.labels = [mapping_dict[key] for key in orig_single_labels]

        if mode in ["both", "multi"]:
            for label in self.label_matrix_labels:
                if label not in mapping_dict.keys():
                    msg = str(label)
                    msg += " from label_matrix_labels was not specified in mapping dictionary."
                    raise MissingKeyForRelabeling(msg)
            orig_multi_labels = self.label_matrix_labels
            self.label_matrix_labels = [mapping_dict[key] for key in orig_multi_labels]

    def relabel_to_max_source(self):
        """ Relabels labels to the source name of most contributing source.
        """

        self.labels = self.label_matrix_labels[self.label_matrix.values.argmax(axis=1)]

    def sample(self, n_samples, random_seed=None, reindex=True):
        """ Returns n random observations.
        """
        if random_seed is not None:
            random.seed(random_seed)

        if n_samples > self.n_samples:
            n_samples = self.n_samples
        indices = random.sample(self.spectra.index.values.tolist(), n_samples)
        return self.get_indices(np.isin(self.spectra.index, indices), reindex=reindex)

    def shuffle(self, random_seed=None):
        """ Shuffles observations.
        """
        if random_seed is not None:
            random.seed(random_seed)
        indices = random.sample(self.spectra.index.values.tolist(), self.n_samples)
        self.sources = self.sources.iloc[indices, :]
        self.spectra = self.spectra.iloc[indices, :]
        self.collection_information = self.collection_information.iloc[indices, :]

    def split(self, split_ratio=0.2, random_seed=None):
        """ Returns n random observations.
        """
        if random_seed is not None:
            random.seed(random_seed)
            np.random.seed(random_seed)

        indices = self.spectra.index.values
        np.random.shuffle(indices)
        val_mask = np.full(len(indices), False)
        train_mask = np.full(len(indices), True)

        n_val = int(len(indices) * split_ratio)
        val_indices = np.random.choice(np.arange(len(indices)), n_val)
        val_mask[val_indices] = True
        train_mask[val_indices] = False

        train_ss = self.get_indices(train_mask)
        val_ss = self.get_indices(val_mask)

        return train_ss, val_ss

    def get_indices(self, indices, reindex=True):
        """ Returns SampleSet of only the observations defined by "indices".
        """
        if isinstance(indices, pd.core.series.Series):
            indices = indices.values
        samples = copy.copy(self)
        samples.sources = samples.sources.loc[indices, :]
        samples.spectra = samples.spectra.loc[indices, :]
        samples.collection_information = samples.collection_information.loc[indices, :]
        if samples.predictions.shape[0]:
            samples.predictions = samples.predictions[indices]
        if samples.prediction_probas.shape[0]:
            try:
                samples.prediction_probas = samples.prediction_probas.loc[indices, :]
            except AttributeError:
                samples.prediction_probas = samples.prediction_probas[indices, :]

        if samples.features.shape[0]:
            try:
                samples.features = samples.features.loc[indices, :]
            except AttributeError:
                samples.features = samples.features[indices, :]

        if reindex:
            samples.sources.index = np.arange(samples.n_samples)
            samples.spectra.index = np.arange(samples.n_samples)
            samples.collection_information.index = np.arange(samples.n_samples)
        return samples

    def get_features(self, include_spectra=True):
        """get_features getter"""
        if include_spectra:
            if self._features.shape[0] > 0:
                if self._features.ndim == 1:
                    floc = np.expand_dims(self._features, 1)
                else:
                    floc = self._features
                return np.nan_to_num(
                    np.concatenate(
                        (self._spectra, floc), axis=1))
            else:
                return np.nan_to_num(self._spectra.values)
        else:
            return np.nan_to_num(self._features.values)

    def to_energy(self, energy_bin_centers, inbin_values=None):
        """ Transforms specrum into energy units centered at
        "energy_bin_centers"
        """
        orig_counts = self.spectra.sum(axis=1)
        energy_spectra = np.zeros([self.n_samples, len(energy_bin_centers)])
        original_energy_centers = self._get_energy_centers(inbin_values)

        loop_values = zip(original_energy_centers, self.spectra.values)
        for j, (old_centers, old_values) in enumerate(loop_values):
            interp_func = interp1d(
                old_centers,
                old_values,
                kind="linear",
                fill_value=0,
                bounds_error=False
            )
            new_spectrum = interp_func(energy_bin_centers)
            energy_spectra[j, :] = new_spectrum
        # Normalize to maintain total counts
        energy_spectra = energy_spectra.clip(0)
        total_counts = energy_spectra.sum(axis=1)
        total_counts[total_counts <= 0] = 1
        energy_spectra = (energy_spectra / total_counts[:, None]) *\
            orig_counts[:, None]
        self.spectra = pd.DataFrame(energy_spectra).replace(np.nan, 0)

    def _get_energy_centers(self, bins=None):
        """  Returns an array of the energy centers for the calibration of each
        spectra in the SampleSet.
        """
        if bins is None:
            bins = np.linspace(0, 1, self.n_channels)

        energy_centers = self.collection_information.ecal_order_0.values[:, None] + \
            bins * self.collection_information.ecal_order_1.values[:, None] + \
            bins**2 * self.collection_information.ecal_order_2.values[:, None] + \
            bins**3 * self.collection_information.ecal_order_3.values[:, None] + \
            self.collection_information.ecal_low_e.values[:, None] / (1 + 60 * bins)
        return energy_centers

    def _combine_redundant_sources(self):
        """Combines redundant columns of sources DataFrame.
        """
        columns = self._sources.columns.values
        self._sources = self._sources.groupby(columns, axis=1, sort=False).sum()

    def add_spectra(self, spectra, sources, labels, info):
        """Appends the given arguments to their respective dataframes.
        """
        self._spectra = self._spectra.append(pd.DataFrame(spectra), ignore_index=True, sort=True)
        self._sources = self._sources.append(pd.DataFrame(sources), ignore_index=True, sort=True)
        self._sources["label"] = labels
        self._collection_information = self._collection_information.append(
            pd.DataFrame(info),
            ignore_index=True,
            sort=True
        )

    def add_spectrum(self, spectrum, source_values, label):
        """Appends the list-like spectrum to the spectra and sources DataFrame.
        """
        if not isinstance(source_values, dict):
            raise TypeError("'source_values' argument must be of type 'dict'")
        if not all([isinstance(x, str) for x in source_values.keys()]):
            raise TypeError("Keys in 'source_values' argument must all be of type 'str'")
        if not all([isinstance(x, float) for x in source_values.values()]):
            raise TypeError("Values in 'source_values' argument must all be of type 'float'")
        if sum(source_values.values()) != 1.0:
            raise ValueError("Values in 'source_values' argument must sum to '1.0'")
        # TODO: check spectrum type and length against existing spectra rows

        self._spectra = self._spectra.append(pd.Series(spectrum), ignore_index=True)
        self._sources = self._sources.append(source_values, ignore_index=True)
        self._sources.at[self._sources.index[-1], "label"] = label

    def add_info(self, live_time=0.0, snr_target=0.0, snr_estimate=0.0, bg_counts=0, fg_counts=0,
                 bg_counts_expected=0, total_counts=0, sigma=0, ecal_order_0=0,
                 ecal_order_1=0.0, ecal_order_2=0.0, ecal_order_3=0.0, ecal_low_e=0.0,
                 real_time=0.0, occupancy_flag=False, total_neutron_counts=0, tag="",
                 date_time=None, desc="", **kwargs):
        """Appends the provide values to the collection information DataFrame.
        """
        row = {
            "live_time": live_time,
            "snr_target": snr_target,
            "snr_estimate": snr_estimate,
            "bg_counts": bg_counts,
            "fg_counts": fg_counts,
            "bg_counts_expected": bg_counts_expected,
            "total_counts": total_counts,
            "sigma": sigma,
            "ecal_order_0": ecal_order_0,
            "ecal_order_1": ecal_order_1,
            "ecal_order_2": ecal_order_2,
            "ecal_order_3": ecal_order_3,
            "ecal_low_e": ecal_low_e,
            "real_time": real_time,
            "occupancy_flag": occupancy_flag,
            "total_neutron_counts": total_neutron_counts,
            "tag": tag,
            "date-time": date_time,
            "descr": desc
        }
        if kwargs:
            for key, value in kwargs.items():
                row[key] = value
        self._collection_information = self._collection_information.append(row, ignore_index=True)

    @property
    def spectra(self):
        """spectra getter setter"""
        return self._spectra

    @spectra.setter
    def spectra(self, value):
        self._spectra = value

    @property
    def sources(self):
        """sources getter setter"""
        return self._sources

    @sources.setter
    def sources(self, value):
        """sources setter"""
        self._sources = value.replace(np.nan, 0)
        self._combine_redundant_sources()

    @property
    def label_matrix_labels(self):
        """Getter for label matrix labels"""
        return self.label_matrix.columns.values

    @label_matrix_labels.setter
    def label_matrix_labels(self, values):
        """Setter for label matrix labels"""
        columns = copy.deepcopy(list(values))
        columns.append("label")
        self._sources.columns = columns
        self._combine_redundant_sources()

    @property
    def features(self):
        """features getter setter"""
        return self._features

    @features.setter
    def features(self, value):
        self._features = value

    @property
    def detector(self):
        """detector getter setter"""
        return self._detector

    @detector.setter
    def detector(self, value):
        self._detector = value

    @property
    def purpose(self):
        """purpose getter setter"""
        return self._purpose

    @purpose.setter
    def purpose(self, value):
        self._purpose = value

    @property
    def distance(self):
        """distance getter setter"""
        return self._collection_information.distance

    @distance.setter
    def distance(self, value):
        self._collection_information.loc[:, "distance"] = value

    @property
    def collection_information(self):
        """collection_information getter setter"""
        return self._collection_information

    @collection_information.setter
    def collection_information(self, value):
        self._collection_information = value

    @property
    def sensor_information(self):
        """sensor_information getter setter"""
        return self._sensor_information

    @sensor_information.setter
    def sensor_information(self, value):
        self._sensor_information = value

    @property
    def config(self):
        """config getter setter"""
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    @property
    def prediction_probas(self):
        """prediction_probas getter setter"""
        return self._prediction_probas

    @prediction_probas.setter
    def prediction_probas(self, value):
        if isinstance(value, pd.DataFrame):
            self._prediction_probas = value
        else:
            self._prediction_probas = pd.DataFrame(value, columns=self.label_matrix_labels)

    @property
    def snr_estimate(self):
        """snr_estimate getter setter"""
        return self._collection_information.snr_estimate

    @snr_estimate.setter
    def snr_estimate(self, value):
        self._collection_information.loc[:, "snr_estimate"] = value

    @property
    def predictions(self):
        """predictions getter setter"""
        return np.array(self._predictions)

    @predictions.setter
    def predictions(self, value):
        self._predictions = value

    @property
    def measured_or_synthetic(self):
        """measured_or_synthetic getter setter"""
        return self._measured_or_synthetic

    @measured_or_synthetic.setter
    def measured_or_synthetic(self, value):
        self._measured_or_synthetic = value

    @property
    def subtract_background(self):
        """subtract_background getter setter"""
        return self._subtract_background

    @subtract_background.setter
    def subtract_background(self, value):
        self._subtract_background = value

    @property
    def comments(self):
        """comments getter setter"""
        return self._comments

    @comments.setter
    def comments(self, value):
        self._comments = value

    # Repeat the below with the rest of the columns of the collection info table
    @property
    def live_time(self):
        """live_time getter setter"""
        return self._collection_information.live_time

    @live_time.setter
    def live_time(self, value):
        self._collection_information.loc[:, "live_time"] = value

    @property
    def sigma(self):
        """sigma getter setter"""
        return self._collection_information.sigma

    @sigma.setter
    def sigma(self, value):
        self._collection_information.loc[:, "sigma"] = value

    @property
    def total_counts(self):
        """total_counts getter setter"""
        return self._collection_information.total_counts

    @total_counts.setter
    def total_counts(self, value):
        self._collection_information.loc[:, "total_counts"] = value

    @property
    def labels(self):
        """labels getter setter"""
        return self._sources.loc[:, "label"].values

    @labels.setter
    def labels(self, value):
        self._sources.loc[:, "label"] = value

    @property
    def label_matrix(self):
        """label_matrix getter"""
        return self._sources.loc[:, self._sources.columns != "label"].replace(np.nan, 0)

    @property
    def n_samples(self):
        """n_samples getter"""
        return self._spectra.shape[0]

    @property
    def n_channels(self):
        """n_channels getter"""
        return self._spectra.shape[1]

    @property
    def n_features(self, include_spectra=True):
        """n_features getter"""
        if include_spectra:
            return self.n_channels + self._features.shape[1]
        else:
            return self._features.shape[1]

    @property
    def source_types(self):
        """source_types getter"""
        return self.label_matrix.columns.values

    @property
    def energy_bin_centers(self):
        """energy_bin_centers getter"""
        if not len(self._energy_bin_centers):
            self._energy_bin_centers = self._get_energy_centers()
        return self._energy_bin_centers

    @energy_bin_centers.setter
    def energy_bin_centers(self, value):
        """energy_bin_centers setter"""
        self._energy_bin_centers = value

    @property
    def ecal_order_0(self):
        """ecal_order_0 getter"""
        return self.collection_information.ecal_order_0

    @ecal_order_0.setter
    def ecal_order_0(self, value):
        """ecal_order_0 setter"""
        self.collection_information.loc[:, "ecal_order_0"] = value

    @property
    def ecal_order_1(self):
        """ecal_order_1 getter"""
        return self.collection_information.ecal_order_1

    @ecal_order_1.setter
    def ecal_order_1(self, value):
        """ecal_order_1 setter"""
        self.collection_information.loc[:, "ecal_order_1"] = value

    @property
    def ecal_order_2(self):
        """ecal_order_2 getter"""
        return self.collection_information.ecal_order_2

    @ecal_order_2.setter
    def ecal_order_2(self, value):
        """ecal_order_2 setter"""
        self.collection_information.loc[:, "ecal_order_2"] = value

    @property
    def ecal_order_3(self):
        """ecal_order_3 getter"""
        return self.collection_information.ecal_order_3

    @ecal_order_3.setter
    def ecal_order_3(self, value):
        """ecal_order_3 setter"""
        self.collection_information.loc[:, "ecal_order_3"] = value

    @property
    def ecal_low_e(self):
        """ecal_low_e getter"""
        return self.collection_information.ecal_low_e

    @ecal_low_e.setter
    def ecal_low_e(self, value):
        """ecal_low_e setter"""
        self.collection_information.loc[:, "ecal_low_e"] = value

    @property
    def ecal_factors(self):
        """ecal_factors getter
        returns calibration factors in following order:
            order_0, order_1, order_2, order_3, low_e
        """
        calibration_factors = np.array([
            self.collection_information.ecal_order_0.values,
            self.collection_information.ecal_order_1.values,
            self.collection_information.ecal_order_2.values,
            self.collection_information.ecal_order_3.values,
            self.collection_information.ecal_low_e.values
        ])
        return calibration_factors

    @ecal_factors.setter
    def ecal_factors(self, calibration_factors):
        """ecal_low_e setter"""
        self.collection_information.loc[:, "ecal_order_0"] = calibration_factors[0]
        self.collection_information.loc[:, "ecal_order_1"] = calibration_factors[1]
        self.collection_information.loc[:, "ecal_order_2"] = calibration_factors[2]
        self.collection_information.loc[:, "ecal_order_3"] = calibration_factors[3]
        self.collection_information.loc[:, "ecal_low_e"] = calibration_factors[4]

    @property
    def neutron_counts(self):
        """ neutron counts for collection
        """
        return self.collection_information.total_neutron_counts

    @neutron_counts.setter
    def neutron_counts(self, neutron_counts):
        """ setter for neutron counts
        """
        self.collection_information.total_neutron_counts = neutron_counts


class RebinningCalculationError(Exception):
    pass


class MissingKeyForRelabeling(Exception):
    pass



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module sets up shared logging functionality and global variables."""
import logging
import os
import sys


HANDLER = logging.StreamHandler(sys.stdout)
logging.root.addHandler(HANDLER)
logging.root.setLevel(logging.DEBUG)
MPL_LOGGER = logging.getLogger('matplotlib')
MPL_LOGGER.setLevel(logging.WARNING)

SMPL_FILE_EXTENSION = ".smpl"
PYRIID_DATA_DIR_ENV_KEY = "PYRIID_DATA_DIR"
DATA_PATH = ""
# Check for required environment variables relating to data directory
if PYRIID_DATA_DIR_ENV_KEY in os.environ:
    DATA_PATH = os.environ[PYRIID_DATA_DIR_ENV_KEY]
else:
    msg = "The following '{}' environment variable is not defined.  ".format(
        PYRIID_DATA_DIR_ENV_KEY
    )
    msg += "Note that certain, optional functions will not work."
    logging.warn(msg)

DATA_PATH = os.path.expanduser(DATA_PATH)


class DataDirectoryNotFoundError(Exception):
    def __init__(self, *args, **kwargs):
        default_message = "The path to the data directory does not exist."
        if not (args or kwargs):
            args = (default_message,)
        super().__init__(*args, **kwargs)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""Simple setup script for installing the core package.
"""
from os import path
from setuptools import find_packages, setup


p = path.abspath(path.dirname(__file__))
with open(path.join(p, './README.md')) as f:
    README = f.read()

setup(
    name="riid",
    version="1.0.1",
    description="Machine learning-based models and utilities for radioisotope identification",
    long_description=README,
    long_description_content_type='text/markdown',
    author="Tyler Morrow,Nathan Price",
    author_email="tmorro@sandia.gov,njprice@sandia.gov",
    url="https://github.com/sandialabs/PyRIID",
    packages=find_packages(),
    python_requires=">=3.7, <3.8",
    install_requires=[
        "tensorflow==2.0.0",
        "tensorflow-model-optimization==0.1.3",
        "numpy==1.17.4",
        "pandas==1.0.5",
        "matplotlib==3.1.2",
        "scikit-learn==0.22",
        "tables==3.6.1",
        "tqdm",
        "seaborn==0.10.1",
        "h5py<3.0.0",
    ],
    # PyPI package information.
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7'
    ],
    license='BSD-3',
    keywords='pyriid riid machine learning',
)



# Copyright 2021 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
"""This module provides custom metrics."""
import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K


def multi_f1(y_true: np.ndarray, y_pred: np.ndarray) -> tf.Tensor:
    """This metric provides a measure of the F1 score of two tensors.

        y_true and y_pred are assumed to sum to 1.

        Args:
            y_true: a list of ground truth.
            y_pred: a list of predictions to compare against the ground truth.

        Returns:
            Returns the custom loss score.

        Raises:
            None.
    """
    diff = y_true - y_pred
    negs = K.clip(diff, -1.0, 0.0)
    false_positive = -K.sum(negs, axis=-1)
    true_positive = 1.0 - false_positive

    return K.mean(true_positive)


def single_f1(y_true: np.ndarray, y_pred: np.ndarray) -> tf.Tensor:
    """ Computes the weighted F1 score for the maximum prediction and maximum
    ground truth.

        y_true and y_pred are assumed to sum to 1.

        Args:
            y_true: a list of ground truth.
            y_pred: a list of predictions to compare against the ground truth.

        Returns:
            Returns the custom loss score.

        Raises:
            None
    """
    a = tf.dtypes.cast(y_true == K.max(y_true, axis=1)[:, None], tf.float32)
    b = tf.dtypes.cast(y_pred == K.max(y_pred, axis=1)[:, None], tf.float32)

    TP_mat = tf.dtypes.cast(K.all(tf.stack([a, b]), axis=0), tf.float32)
    FP_mat = tf.dtypes.cast(K.all(tf.stack([a != b, b == 1]), axis=0), tf.float32)
    FN_mat = tf.dtypes.cast(K.all(tf.stack([a != b, a == 1]), axis=0), tf.float32)

    TPs = K.sum(TP_mat, axis=0)
    FPs = K.sum(FP_mat, axis=0)
    FNs = K.sum(FN_mat, axis=0)

    F1s = 2 * TPs / (2*TPs + FNs + FPs + tf.fill(tf.shape(TPs), K.epsilon()))

    support = K.sum(a, axis=0)
    f1 = K.sum(F1s * support) / K.sum(support)
    return f1
